from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Iterable, Optional, Text

import structlog
from pydantic import ValidationError

import rasa.shared
from rasa.core.brokers.broker import EventBroker
from rasa.core.iam_credentials_providers.credentials_provider_protocol import (
    SupportedServiceType,
)
from rasa.core.redis_connection_factory import (
    DeploymentMode,
    RedisConfig,
    RedisConnectionFactory,
)
from rasa.core.tracker_stores.tracker_store import SerializedTrackerAsText, TrackerStore
from rasa.shared.core.domain import Domain
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException

structlogger = structlog.get_logger(__name__)

# default value for key prefix in RedisTrackerStore
DEFAULT_REDIS_TRACKER_STORE_KEY_PREFIX = "tracker:"


class RedisTrackerStore(TrackerStore, SerializedTrackerAsText):
    """Stores conversation history in Redis."""

    def __init__(
        self,
        domain: Domain,
        host: Text = "localhost",
        port: int = 6379,
        db: int = 0,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        event_broker: Optional[EventBroker] = None,
        record_exp: Optional[float] = None,
        key_prefix: Optional[Text] = None,
        use_ssl: bool = False,
        ssl_keyfile: Optional[Text] = None,
        ssl_certfile: Optional[Text] = None,
        ssl_ca_certs: Optional[Text] = None,
        deployment_mode: Text = DeploymentMode.STANDARD.value,
        endpoints: Optional[list] = None,
        sentinel_service: Optional[Text] = None,
        **kwargs: Dict[Text, Any],
    ) -> None:
        """Initializes the tracker store."""

        # Create Redis connection using the factory directly
        try:
            config = RedisConfig(
                host=host,
                port=port,
                service_type=SupportedServiceType.TRACKER_STORE,
                db=db,
                username=username,
                password=password,
                use_ssl=use_ssl,
                ssl_keyfile=ssl_keyfile,
                ssl_certfile=ssl_certfile,
                ssl_ca_certs=ssl_ca_certs,
                deployment_mode=deployment_mode,
                endpoints=endpoints,
                sentinel_service=sentinel_service,
                decode_responses=True,
            )
            self.red = RedisConnectionFactory.create_connection(config)
        except ValidationError as e:
            raise RasaException(f"Invalid Redis configuration: {e}")

        self.record_exp = record_exp

        self.key_prefix = DEFAULT_REDIS_TRACKER_STORE_KEY_PREFIX
        if key_prefix:
            structlogger.debug(
                "redis_tracker_store.init.custom_key_prefix",
                event_info=f"Setting non-default redis key prefix: '{key_prefix}'.",
            )
            self._set_key_prefix(key_prefix)

        super().__init__(domain, event_broker, **kwargs)

    def _set_key_prefix(self, key_prefix: Text) -> None:
        if isinstance(key_prefix, str) and key_prefix.isalnum():
            self.key_prefix = key_prefix + ":" + DEFAULT_REDIS_TRACKER_STORE_KEY_PREFIX
        else:
            structlogger.warning(
                "redis_tracker_store.init.invalid_key_prefix",
                event_info=(
                    f"Omitting provided non-alphanumeric "
                    f"redis key prefix: '{key_prefix}'. "
                    f"Using default '{self.key_prefix}' instead."
                ),
            )

    def _get_key_prefix(self) -> Text:
        return self.key_prefix

    async def save(
        self, tracker: DialogueStateTracker, timeout: Optional[float] = None
    ) -> None:
        """Saves the current conversation state."""
        await self.stream_events(tracker)

        if not timeout and self.record_exp:
            timeout = self.record_exp

        # if the sender_id starts with the key prefix, we remove it
        # this is used to avoid storing the prefix twice
        sender_id = tracker.sender_id
        if sender_id.startswith(self.key_prefix):
            sender_id = sender_id[len(self.key_prefix) :]

        stored = self.red.get(self.key_prefix + sender_id)

        if stored is not None:
            prior_tracker = self.deserialise_tracker(sender_id, stored)

            tracker = self._merge_trackers(prior_tracker, tracker)

        serialised_tracker = self.serialise_tracker(tracker)
        self.red.set(self.key_prefix + sender_id, serialised_tracker, ex=timeout)

    async def delete(self, sender_id: Text) -> None:
        """Delete tracker for the given sender_id.

        Args:
            sender_id: Sender id of the tracker to be deleted.
        """
        if not await self.exists(sender_id):
            structlogger.info(
                "redis_tracker_store.delete.no_tracker_for_sender_id",
                event_info=f"Could not find tracker for conversation ID '{sender_id}'.",
            )
            return None

        if sender_id.startswith(self.key_prefix):
            sender_id = sender_id[len(self.key_prefix) :]

        self.red.delete(self.key_prefix + sender_id)
        structlogger.info(
            "redis_tracker_store.delete.deleted_tracker",
            sender_id=sender_id,
        )

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Retrieves tracker for the latest conversation session.

        The Redis key is formed by appending a prefix to sender_id.

        Args:
            sender_id: Conversation ID to fetch the tracker for.

        Returns:
            Tracker containing events from the latest conversation sessions.
        """
        return await self._retrieve(sender_id, fetch_all_sessions=False)

    async def retrieve_full_tracker(
        self, sender_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Retrieves tracker for all conversation sessions.

        The Redis key is formed by appending a prefix to sender_id.

        Args:
            sender_id: Conversation ID to fetch the tracker for.

        Returns:
            Tracker containing events from all conversation sessions.
        """
        return await self._retrieve(sender_id, fetch_all_sessions=True)

    async def _retrieve(
        self, sender_id: Text, fetch_all_sessions: bool
    ) -> Optional[DialogueStateTracker]:
        """Returns tracker matching sender_id.

        Args:
            sender_id: Conversation ID to fetch the tracker for.
            fetch_all_sessions: Whether to fetch all sessions or only the last one.
        """
        if sender_id.startswith(self.key_prefix):
            sender_id = sender_id[len(self.key_prefix) :]

        stored = self.red.get(self.key_prefix + sender_id)
        if stored is None:
            structlogger.debug(
                "redis_tracker_store.retrieve.no_tracker_for_sender_id",
                event_info=f"Could not find tracker for conversation ID '{sender_id}'.",
            )
            return None

        tracker = self.deserialise_tracker(sender_id, stored)
        if fetch_all_sessions:
            return tracker

        # only return the last session
        multiple_tracker_sessions = (
            rasa.shared.core.trackers.get_trackers_for_conversation_sessions(tracker)
        )

        if len(multiple_tracker_sessions) <= 1:
            return tracker

        return multiple_tracker_sessions[-1]

    async def keys(self) -> Iterable[Text]:
        """Returns keys of the Redis Tracker Store."""
        return self.red.keys(self.key_prefix + "*")

    @staticmethod
    def _merge_trackers(
        prior_tracker: DialogueStateTracker, tracker: DialogueStateTracker
    ) -> DialogueStateTracker:
        """Merges two trackers.

        Args:
            prior_tracker: Tracker containing events from the previous conversation
                sessions.
            tracker: Tracker containing events from the current conversation session.
        """
        if not prior_tracker.events:
            return tracker

        last_event_timestamp = prior_tracker.events[-1].timestamp
        past_tracker = tracker.travel_back_in_time(target_time=last_event_timestamp)

        if past_tracker.events == prior_tracker.events:
            return tracker

        merged = tracker.init_copy()
        merged.update_with_events(list(prior_tracker.events), override_timestamp=False)

        for new_event in tracker.events:
            # Event subclasses implement `__eq__` method that make it difficult
            # to compare events. We use `as_dict` to compare events.
            if all(
                [
                    new_event.as_dict() != existing_event.as_dict()
                    for existing_event in merged.events
                ]
            ):
                merged.update(new_event)

        return merged

    async def update(
        self, tracker: DialogueStateTracker, apply_deletion_only: bool = True
    ) -> None:
        """Overwrites the tracker for the given sender_id."""
        serialised_tracker = self.serialise_tracker(tracker)

        # if the sender_id starts with the key prefix, we remove it
        # this is used to avoid storing the prefix twice
        sender_id = tracker.sender_id
        if sender_id.startswith(self.key_prefix):
            sender_id = sender_id[len(self.key_prefix) :]

        self.red.set(
            self.key_prefix + sender_id, serialised_tracker, ex=self.record_exp
        )

        first_event_timestamp = str(datetime.fromtimestamp(tracker.events[0].timestamp))

        structlogger.info(
            "redis_tracker_store.update.updated_tracker",
            sender_id=tracker.sender_id,
            first_event_timestamp=first_event_timestamp,
        )



import random
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Protocol, Set, Tuple, Type

import structlog

from rasa.dialogue_understanding.commands.prompt_command import PromptCommand
from rasa.e2e_test.e2e_test_case import TestSuite
from rasa.llm_fine_tuning.llm_data_preparation_module import LLMDataExample
from rasa.llm_fine_tuning.storage import StorageContext
from rasa.llm_fine_tuning.utils import commands_as_string

TRAIN_TEST_MODULE_STORAGE_LOCATION = "4_train_test_split"

INSTRUCTION_DATA_FORMAT = "instruction"
CONVERSATIONAL_DATA_FORMAT = "conversational"

KEY_COMMANDS = "commands"
KEY_DATA_EXAMPLES = "data_examples"
KEY_TEST_CASE_NAME = "test_case_name"

structlogger = structlog.get_logger()


# A protocol (similar to an interface) for data example that all fine-tuning data
# format classes should implement.
class DataExampleFormat(Protocol):
    def as_dict(self) -> dict: ...


@dataclass
class InstructionDataFormat(DataExampleFormat):
    # The prompt and completion fields are used to store the input and output of the
    # LLM model. The instruction data format is taken from the TRL library.
    # Refer: https://huggingface.co/docs/trl/en/sft_trainer#dataset-format-support
    # The data format can also be used to train on Azure. Refer:
    # https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/fine-tuning

    prompt: str
    completion: str

    def as_dict(self) -> Dict[str, str]:
        return {"prompt": self.prompt, "completion": self.completion}


@dataclass
class ConversationalMessageDataFormat:
    role: str
    content: str


@dataclass
class ConversationalDataFormat(DataExampleFormat):
    # The conversation data format is taken from the TRL library.
    # Refer: https://huggingface.co/docs/trl/en/sft_trainer#dataset-format-support
    # The data format can also be used to train on Azure. Refer:
    # https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/fine-tuning

    messages: List[ConversationalMessageDataFormat]

    def as_dict(self) -> Dict[str, List[Dict[str, str]]]:
        return {
            "messages": [
                {"role": message.role, "content": message.content}
                for message in self.messages
            ]
        }


def _get_command_types_covered_by_llm_data_point(
    data_point: LLMDataExample,
) -> Set[Type[PromptCommand]]:
    """Get the command types covered by the LLM data point.

    This function returns the set of command types from the output present in a
    LLMDataExample object. Eg: The function returns {'SetSlotCommand',
    'StartFlowCommand'} when the LLMDataExample.output is 'SetSlotCommand(slot, abc),
    SetSlotCommand(slot, cde), StartFlowCommand(xyz)'.
    """
    commands_covered = set()
    for command in data_point.output:
        commands_covered.add(command.__class__)
    return commands_covered


def _get_grouped_fine_tuning_data_by_test_name_with_commands(
    fine_tuning_data: List[LLMDataExample],
) -> List[Dict[str, Any]]:
    """Group the fine-tune data by test name and capture commands covered by each group.

    This function groups the fine-tuning data by original_test_name and captures the
    commands covered by each group. The function returns a dictionary where the keys
    are the original_test_name and the values are dictionaries containing the data
    and commands covered by each group.
    """
    grouped_data: Dict[str, Dict[str, Any]] = defaultdict(
        lambda: {KEY_DATA_EXAMPLES: [], KEY_COMMANDS: set()}
    )

    for data_point in fine_tuning_data:
        grouped_data[data_point.original_test_name][KEY_DATA_EXAMPLES].append(
            data_point
        )
        command_types = _get_command_types_covered_by_llm_data_point(data_point)
        for command_type in command_types:
            grouped_data[data_point.original_test_name][KEY_COMMANDS].add(command_type)

    # Convert the grouped data to the required format.
    result = [
        {
            KEY_TEST_CASE_NAME: test_case_name,
            KEY_DATA_EXAMPLES: group[KEY_DATA_EXAMPLES],
            KEY_COMMANDS: group[KEY_COMMANDS],
        }
        for test_case_name, group in grouped_data.items()
    ]

    return result


def _get_minimum_test_case_groups_to_cover_all_commands(
    grouped_data: List[Dict[str, Any]],
) -> List[str]:
    """Get the minimum test case groups to cover all commands.

    This function implements a greedy algorithm that returns the minimum test case
    groups to cover all commands. The function takes a list of dictionaries where each
    dictionary represents a test group and contains the test_case_name, data_examples,
    and commands covered by each group.

    The function iteratively selects the test case group that covers the most number
    of uncovered commands until all commands are covered.

    The function returns the selected test case groups as a list of strings.

    Example:
    data = [
        {
            "test_case_name": "t1",
            "data_examples": [],
            "commands": {"SetSlotCommand", "CancelFlowCommand"}
        },
        {
            "test_case_name": "t2",
            "data_examples": [],
            "commands": {"CancelFlowCommand"}
        },
        {"test_case_name": "t3", "data_examples": [], "commands": {"StartFlowCommand"}},
        {
            "test_case_name": "t4",
            "data_examples": [],
            "commands": {"SetSlotCommand", "StartFlowCommand"}
        },
    ]

    selected_test_cases = _get_minimum_test_case_groups_to_cover_all_commands(data)
    print(selected_test_cases)
    # Output: ['t1', 't3']
    """
    # Get all the commands covered from the finetuning data.
    all_commands = set(
        command for test_group in grouped_data for command in test_group[KEY_COMMANDS]
    )
    selected_test_cases = []
    covered_commands: Set[Type[PromptCommand]] = set()

    while covered_commands != all_commands:
        # Find the test case group that covers the most number of uncovered commands
        best_test_case_group = None
        commands_covered_by_best_test_case_group: Set[str] = set()

        for test_group in grouped_data:
            commands_covered = test_group[KEY_COMMANDS] - covered_commands
            if len(commands_covered) > len(commands_covered_by_best_test_case_group):
                best_test_case_group = test_group[KEY_TEST_CASE_NAME]
                commands_covered_by_best_test_case_group = commands_covered

        if best_test_case_group is None:
            break

        selected_test_cases.append(best_test_case_group)
        covered_commands.update(commands_covered_by_best_test_case_group)

    structlogger.info(
        "llm_fine_tuning.train_test_split_module.command_coverage_in_train_dataset",
        covered_commands=[command.__name__ for command in covered_commands],
    )
    return selected_test_cases


def _get_finetuning_data_in_instruction_data_format(
    train_data: List[Dict[str, Any]], validation_data: List[Dict[str, Any]]
) -> Tuple[List[DataExampleFormat], List[DataExampleFormat]]:
    """Convert the fine-tuning data to the required Alpaca format.

    The function takes the train and validation data as input and returns the formatted
    train and validation data.
    """

    def _convert_into_instruction_data_format(
        data: List[Dict[str, Any]],
    ) -> List[DataExampleFormat]:
        return [
            InstructionDataFormat(
                llm_data_example.prompt,
                commands_as_string(llm_data_example.output),
            )
            for test_group in data
            for llm_data_example in test_group[KEY_DATA_EXAMPLES]
        ]

    formatted_train_data = _convert_into_instruction_data_format(train_data)
    formatted_validation_data = _convert_into_instruction_data_format(validation_data)
    return formatted_train_data, formatted_validation_data


def _get_finetuning_data_in_conversational_data_format(
    train_data: List[Dict[str, Any]], validation_data: List[Dict[str, Any]]
) -> Tuple[List[DataExampleFormat], List[DataExampleFormat]]:
    """Convert the fine-tuning data to the required ShareGPT format.

    The function takes the train and validation data as input and returns the formatted
    train and validation data.
    """

    def _convert_into_conversational_data_format(
        data: List[Dict[str, Any]],
    ) -> List[DataExampleFormat]:
        return [
            ConversationalDataFormat(
                [
                    ConversationalMessageDataFormat("user", llm_data_example.prompt),
                    ConversationalMessageDataFormat(
                        "assistant", commands_as_string(llm_data_example.output)
                    ),
                ]
            )
            for test_group in data
            for llm_data_example in test_group[KEY_DATA_EXAMPLES]
        ]

    formatted_train_data = _convert_into_conversational_data_format(train_data)
    formatted_validation_data = _convert_into_conversational_data_format(
        validation_data
    )
    return formatted_train_data, formatted_validation_data


def _check_and_log_missing_validation_dataset_command_coverage(
    grouped_data: List[Dict[str, Any]],
    validation_data: List[Dict[str, Any]],
) -> None:
    """Check and log the missing commands in the validation data.

    Args:
        grouped_data: The grouped fine-tuning data - complete dataset.
        validation_data: The validation data to check for missing commands.
    """
    all_commands = set(
        command for test_group in grouped_data for command in test_group[KEY_COMMANDS]
    )
    command_coverage_in_validation_dataset = set(
        command
        for test_group in validation_data
        for command in test_group[KEY_COMMANDS]
    )

    missing_commands = all_commands - command_coverage_in_validation_dataset
    if missing_commands:
        structlogger.warning(
            "llm_fine_tuning.train_test_split_module.missing_commands_in_validation_dat"
            "aset",
            missing_commands=[command.__name__ for command in missing_commands],
        )


def _get_train_validation_data_splits(
    fine_tuning_data: List[LLMDataExample], train_frac: float
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
    # Group by original_test_name and capture commands covered by each group.
    grouped_data = _get_grouped_fine_tuning_data_by_test_name_with_commands(
        fine_tuning_data
    )

    # Get the minimum test case groups to cover all commands. This is done to ensure
    # that the training data covers all the commands present in the fine-tuning data.
    selected_tests = _get_minimum_test_case_groups_to_cover_all_commands(grouped_data)

    # Separate the tests into train data and remaining data.
    train_data = [
        data for data in grouped_data if data[KEY_TEST_CASE_NAME] in selected_tests
    ]
    remaining_data = [
        data for data in grouped_data if data[KEY_TEST_CASE_NAME] not in selected_tests
    ]

    # Shuffle the remaining data randomly.
    random.shuffle(remaining_data)

    # Calculate the number of samples required for training
    total_samples = len(grouped_data)
    train_size = int(total_samples * train_frac)

    # Number of additional samples needed in train set
    current_train_size = len(train_data)
    additional_train_needed = train_size - current_train_size

    if additional_train_needed > 0:
        # Select additional samples from the remaining data
        additional_train_samples = remaining_data[:additional_train_needed]
        train_data.extend(additional_train_samples)

        # Update remaining data after adding to train data
        remaining_data = remaining_data[additional_train_needed:]

    # As the train dataset is selected based on the commands covered, the remaining data
    # might not have all the commands. So, we need check and throw a warning if any
    # command is missing in the remaining data. Ensuring command coverage over the
    # validation data will be handled in https://rasahq.atlassian.net/browse/ENG-1217.
    _check_and_log_missing_validation_dataset_command_coverage(
        grouped_data, remaining_data
    )
    return train_data, remaining_data


def split_e2e_test_suite(
    e2e_test_suite: TestSuite,
    train_data: List[Dict[str, Any]],
    validation_data: List[Dict[str, Any]],
) -> Tuple[TestSuite, TestSuite]:
    """Split the e2e test suite into train and validation test suites.

    Args:
        e2e_test_suite: The e2e test suite to split.
        train_data: The train data is used to get the names of the tests that should be
            included in the train test suite from the e2e test suite.
        validation_data: The validation data is used to get the names of the tests that
            should be included in the validation test suite from the e2e test suite.

    Returns:
        Tuple[TestSuite, TestSuite]: The train and validation test suites.
    """
    # Get the names of the test cases that should be included in the train and
    # validation test suites.
    train_test_case_names = {test[KEY_TEST_CASE_NAME] for test in train_data}
    validation_test_case_names = {test[KEY_TEST_CASE_NAME] for test in validation_data}

    train_test_cases = []
    validation_test_cases = []

    for test_case in e2e_test_suite.test_cases:
        if f"{test_case.file}::{test_case.name}" in train_test_case_names:
            train_test_cases.append(test_case)
        elif f"{test_case.file}::{test_case.name}" in validation_test_case_names:
            validation_test_cases.append(test_case)

    # Creating new TestSuite instances for train and validation
    train_suite = TestSuite(
        test_cases=train_test_cases,
        fixtures_per_test=e2e_test_suite.fixtures_per_test,
        metadata=e2e_test_suite.metadata,
        stub_custom_actions=e2e_test_suite.stub_custom_actions,
    )
    validation_suite = TestSuite(
        test_cases=validation_test_cases,
        fixtures_per_test=e2e_test_suite.fixtures_per_test,
        metadata=e2e_test_suite.metadata,
        stub_custom_actions=e2e_test_suite.stub_custom_actions,
    )

    return train_suite, validation_suite


def split_llm_fine_tuning_data(
    fine_tuning_data: List[LLMDataExample],
    train_frac: float,
    output_format: str,
    storage_context: StorageContext,
    e2e_test_suite: TestSuite,
) -> Tuple[List[DataExampleFormat], List[DataExampleFormat]]:
    # Split the fine-tuning data into train and validation data.
    train_data, validation_data = _get_train_validation_data_splits(
        fine_tuning_data, train_frac
    )
    if not validation_data:
        structlogger.warning(
            "llm_fine_tuning.train_test_split_module.empty_validation_dataset",
            event_info=(
                "Validation data is empty. Please provide more data to split into"
                " train and validation data. Also, check if the --train-frac value is"
                " appropriate."
            ),
            train_frac=train_frac,
        )

    # Convert the fine-tuning data to the required format.
    if output_format == INSTRUCTION_DATA_FORMAT:
        formatted_train_data, formatted_validation_data = (
            _get_finetuning_data_in_instruction_data_format(train_data, validation_data)
        )
    elif output_format == CONVERSATIONAL_DATA_FORMAT:
        formatted_train_data, formatted_validation_data = (
            _get_finetuning_data_in_conversational_data_format(
                train_data, validation_data
            )
        )
    else:
        raise ValueError(
            f"Output format '{output_format}' is not supported. Supported formats are "
            f"'{INSTRUCTION_DATA_FORMAT}' and '{CONVERSATIONAL_DATA_FORMAT}'."
        )

    # Write the train and validation data to the storage location.
    storage_context.write_formatted_finetuning_data(
        formatted_train_data,
        TRAIN_TEST_MODULE_STORAGE_LOCATION,
        "ft_splits/train.jsonl",
    )
    storage_context.write_formatted_finetuning_data(
        formatted_validation_data,
        TRAIN_TEST_MODULE_STORAGE_LOCATION,
        "ft_splits/val.jsonl",
    )

    # Split the e2e test suite data into train and validation data.
    train_suite, validation_suite = split_e2e_test_suite(
        e2e_test_suite, train_data, validation_data
    )

    # Write the train and validation test suite data to the storage location.
    storage_context.write_e2e_test_suite_to_yaml_file(
        train_suite, TRAIN_TEST_MODULE_STORAGE_LOCATION, "e2e_tests/train.yaml"
    )
    storage_context.write_e2e_test_suite_to_yaml_file(
        validation_suite,
        TRAIN_TEST_MODULE_STORAGE_LOCATION,
        "e2e_tests/validation.yaml",
    )

    return formatted_train_data, formatted_validation_data



from __future__ import annotations

import abc
import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, Text

from pydantic import BaseModel

import rasa
from rasa.core.actions.action_exceptions import DomainNotFound
from rasa.core.actions.constants import DEFAULT_SELECTIVE_DOMAIN, SELECTIVE_DOMAIN
from rasa.shared.constants import DOCS_BASE_URL
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

if TYPE_CHECKING:
    from rasa.shared.core.domain import Domain
    from rasa.shared.core.trackers import DialogueStateTracker


logger = logging.getLogger(__name__)


class ActionResultType(Enum):
    SUCCESS = "success"
    RETRY_WITH_DOMAIN = "retry_with_domain"


class ActionResult(BaseModel):
    """Result of custom action execution.

    This is used to avoid raising exceptions for expected conditions
    like missing domain (449 status code), which would otherwise be
    captured by tracing as errors.
    """

    result_type: ActionResultType
    response: Optional[Dict[Text, Any]] = None


class CustomActionExecutor(abc.ABC):
    """Interface for custom action executors.

    Provides an abstraction layer for executing custom actions
    regardless of the communication protocol.
    """

    @abc.abstractmethod
    async def run(
        self,
        tracker: "DialogueStateTracker",
        domain: "Domain",
        include_domain: bool = False,
    ) -> Dict[Text, Any]:
        """Executes the custom action.

        Args:
            tracker: The current state of the dialogue.
            domain: The domain object containing domain-specific information.
            include_domain: If True, the domain is included in the request.

        Returns:
            The response from the execution of the custom action.
        """
        pass

    async def run_with_result(
        self,
        tracker: "DialogueStateTracker",
        domain: "Domain",
        include_domain: bool = False,
    ) -> ActionResult:
        """Executes the custom action and returns a result.

        This method is used to avoid raising exceptions for expected conditions
        like missing domain, which would otherwise be captured by tracing as errors.

        By default, this method calls the run method and wraps the response
        for backward compatibility.

        Args:
            tracker: The current state of the dialogue.
            domain: The domain object containing domain-specific information.
            include_domain: If True, the domain is included in the request.

        Returns:
            ActionResult containing the response and result type.
        """
        try:
            response = await self.run(tracker, domain, include_domain)
            return ActionResult(result_type=ActionResultType.SUCCESS, response=response)
        except DomainNotFound:
            return ActionResult(result_type=ActionResultType.RETRY_WITH_DOMAIN)


class NoEndpointCustomActionExecutor(CustomActionExecutor):
    """Implementation of a custom action executor when endpoint is not set.

    Used to handle the case where no endpoint is configured.

    Raises RasaException when executed.
    """

    def __init__(self, action_name: str) -> None:
        """Initializes the custom action executor.

        Args:
            action_name: The name of the custom action.
        """
        self.action_name = action_name

    async def run(
        self,
        tracker: "DialogueStateTracker",
        domain: "Domain",
        include_domain: bool = False,
    ) -> Dict[Text, Any]:
        """Executes the custom action.

        Args:
            tracker: The current state of the dialogue.
            domain: The domain object containing domain-specific information.
            include_domain: If True, the domain is included in the request.

        Returns:
            The response from the execution of the custom action.
        """
        raise RasaException(
            f"Failed to execute custom action '{self.action_name}' "
            f"because no endpoint is configured to run this "
            f"custom action. Please take a look at "
            f"the docs and set an endpoint configuration via the "
            f"--endpoints flag. "
            f"{DOCS_BASE_URL}/action-server/actions"
        )


class CustomActionRequestWriter:
    """Writes the request payload for a custom action."""

    def __init__(self, action_name: str, action_endpoint: EndpointConfig) -> None:
        """Initializes the request writer.

        Args:
            action_name: The name of the custom action.
            action_endpoint: The endpoint configuration for the action server.
        """
        self.action_name = action_name
        self.action_endpoint = action_endpoint

    def _is_selective_domain_enabled(self) -> bool:
        """Check if selective domain handling is enabled.

        Returns:
            True if selective domain handling is enabled, otherwise False.
        """
        if self.action_endpoint is None:
            return False
        return bool(
            self.action_endpoint.kwargs.get(SELECTIVE_DOMAIN, DEFAULT_SELECTIVE_DOMAIN)
        )

    def create(
        self,
        tracker: "DialogueStateTracker",
        domain: "Domain",
        include_domain: bool = False,
    ) -> Dict[str, Any]:
        """Create the JSON payload for the action server request.

        Args:
            tracker: The current state of the dialogue.
            domain: The domain object containing domain-specific information.
            include_domain: If True, the domain is included in the request.

        Returns:
            A JSON payload to be sent to the action server.
        """
        from rasa.shared.core.trackers import EventVerbosity

        tracker_state = tracker.current_state(EventVerbosity.ALL)

        result = {
            "next_action": self.action_name,
            "sender_id": tracker.sender_id,
            "tracker": tracker_state,
            "version": rasa.__version__,
        }

        if include_domain and (
            not self._is_selective_domain_enabled()
            or domain.does_custom_action_explicitly_need_domain(self.action_name)
        ):
            result["domain"] = domain.as_dict()

        result["domain_digest"] = domain.fingerprint()

        return result


class RetryCustomActionExecutor(CustomActionExecutor):
    """Retries the execution of a custom action."""

    def __init__(self, custom_action_executor: CustomActionExecutor) -> None:
        self._custom_action_executor = custom_action_executor

    async def run(
        self,
        tracker: "DialogueStateTracker",
        domain: "Domain",
        include_domain: bool = False,
    ) -> Dict[Text, Any]:
        """Runs the wrapped custom action executor with retry logic.

        First request to the action server is made with/without the domain
        as specified by the `include_domain` parameter.

        If the action server responds with a missing domain indication,
        retries the request with the domain included.

        Args:
            tracker: The current state of the dialogue.
            domain: The domain object containing domain-specific information.
            include_domain: If True, the domain is included in the request.

        Returns:
            The response from the execution of the custom action.

        Raises:
            DomainNotFound: If the action server still requires domain after retry.
        """
        result = await self._custom_action_executor.run_with_result(
            tracker,
            domain,
            include_domain=include_domain,
        )

        if result.result_type == ActionResultType.RETRY_WITH_DOMAIN:
            # Retry with domain included
            result = await self._custom_action_executor.run_with_result(
                tracker, domain, include_domain=True
            )

            # If still missing domain after retry, raise error
            if result.result_type == ActionResultType.RETRY_WITH_DOMAIN:
                raise DomainNotFound()

        return result.response if result.response is not None else {}



from typing import Optional

from rasa.agents.core.types import AgentIdentifier, ProtocolType
from rasa.core.available_agents import AgentConfig, ProtocolConfig
from rasa.shared.core.flows.steps import CallFlowStep


def make_agent_identifier(
    agent_name: str, protocol_type: ProtocolType
) -> AgentIdentifier:
    """Make an agent identifier."""
    return AgentIdentifier(agent_name, protocol_type)


def get_protocol_type(
    step: CallFlowStep, agent_config: Optional[AgentConfig]
) -> ProtocolType:
    """Get the protocol type for an agent.

    Args:
        step: The step that is calling the agent.
        agent_config: The agent configuration.

    Returns:
        The protocol type for the agent.
    """
    if step.exit_if:
        protocol_type = ProtocolType.MCP_TASK
    else:
        protocol_type = (
            ProtocolType.A2A
            if agent_config and agent_config.agent.protocol == ProtocolConfig.A2A
            else ProtocolType.MCP_OPEN
        )
    return protocol_type



import functools
import json
import logging
import re
from json import JSONDecodeError
from pathlib import Path
from re import Match, Pattern
from typing import Any, Dict, List, Optional, Text, Tuple, Union

import structlog

import rasa.shared.data
import rasa.shared.utils.io
from rasa.shared.constants import (
    DOCS_URL_RULES,
    DOCS_URL_SLOTS,
    DOCS_URL_STORIES,
    INTENT_MESSAGE_PREFIX,
    TEST_STORIES_FILE_PREFIX,
)
from rasa.shared.core.constants import LOOP_NAME, RULE_SNIPPET_ACTION_NAME
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import (
    ActiveLoop,
    DialogueStackUpdated,
    SlotSet,
    UserUttered,
)
from rasa.shared.core.slots import ListSlot, TextSlot
from rasa.shared.core.training_data.story_reader.story_reader import StoryReader
from rasa.shared.core.training_data.structures import StoryStep
from rasa.shared.exceptions import YamlException
from rasa.shared.nlu.constants import (
    ACTION_TEXT,
    ENTITIES,
    ENTITY_ATTRIBUTE_END,
    ENTITY_ATTRIBUTE_START,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_VALUE,
    EXTRACTOR,
    FULL_RETRIEVAL_INTENT_NAME_KEY,
    INTENT,
    INTENT_NAME_KEY,
    INTENT_RANKING_KEY,
    PREDICTED_CONFIDENCE_KEY,
    TEXT,
)
from rasa.shared.nlu.training_data import entities_parser
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.utils.yaml import (
    is_key_in_yaml,
    read_yaml,
    validate_raw_yaml_using_schema_file,
    validate_training_data_format_version,
)

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()

KEY_STORIES = "stories"
KEY_STORY_NAME = "story"
KEY_RULES = "rules"
KEY_RULE_NAME = "rule"
KEY_STEPS = "steps"
KEY_ENTITIES = "entities"
KEY_USER_INTENT = "intent"
KEY_USER_MESSAGE = "user"
KEY_COMMANDS = "commands"
KEY_SLOT_NAME = "slot_was_set"
KEY_ACTIVE_LOOP = "active_loop"
KEY_ACTION = "action"
KEY_BOT_END_TO_END_MESSAGE = "bot"
KEY_CHECKPOINT = "checkpoint"
KEY_CHECKPOINT_SLOTS = "slot_was_set"
KEY_METADATA = "metadata"
KEY_OR = "or"
KEY_RULE_CONDITION = "condition"
KEY_WAIT_FOR_USER_INPUT_AFTER_RULE = "wait_for_user_input"
KEY_RULE_FOR_CONVERSATION_START = "conversation_start"
KEY_STACK_UPDATE = "stack"


CORE_SCHEMA_FILE = "shared/utils/schemas/stories.yml"
DEFAULT_VALUE_TEXT_SLOTS = "filled"
DEFAULT_VALUE_LIST_SLOTS = [DEFAULT_VALUE_TEXT_SLOTS]

INTENT_ENTITIES_PATTERN = (
    f"^{INTENT_MESSAGE_PREFIX}"
    f"(?P<{INTENT_NAME_KEY}>[^{{@]+)"  # "{{" is a masked "{" in an f-string
    f"(?P<{PREDICTED_CONFIDENCE_KEY}>@[0-9.]+)?"
    f"(?P<{ENTITIES}>{{.+}})?"  # "{{" is a masked "{" in an f-string
    f"(?P<rest>.*)"
)


class YAMLStoryReader(StoryReader):
    """Class that reads Core training data and rule data in YAML format."""

    @classmethod
    def from_reader(cls, reader: "YAMLStoryReader") -> "YAMLStoryReader":
        """Create a reader from another reader.

        Args:
            reader: Another reader.

        Returns:
            A new reader instance.
        """
        return cls(reader.domain, reader.source_name)

    def read_from_file(
        self, filename: Union[Text, Path], skip_validation: bool = False
    ) -> List[StoryStep]:
        """Read stories or rules from file.

        Args:
            filename: Path to the story/rule file.
            skip_validation: `True` if the file was already validated
                e.g. when it was stored in the database.

        Returns:
            `StoryStep`s read from `filename`.
        """
        self.source_name = str(filename)
        try:
            return self.read_from_string(
                rasa.shared.utils.io.read_file(
                    filename, rasa.shared.utils.io.DEFAULT_ENCODING
                ),
                skip_validation,
            )
        except YamlException as e:
            e.filename = str(filename)
            raise e

    def read_from_string(
        self, string: Text, skip_validation: bool = False
    ) -> List[StoryStep]:
        """Read stories or rules from a string.

        Args:
            string: Unprocessed YAML file content.
            skip_validation: `True` if the string was already validated
                e.g. when it was stored in the database.

        Returns:
            `StoryStep`s read from `string`.
        """
        if not skip_validation:
            validate_raw_yaml_using_schema_file(string, CORE_SCHEMA_FILE)

        yaml_content = read_yaml(string)

        return self.read_from_parsed_yaml(yaml_content)

    def read_from_parsed_yaml(
        self, parsed_content: Dict[Text, Union[Dict, List]]
    ) -> List[StoryStep]:
        """Read stories from parsed YAML.

        Args:
            parsed_content: The parsed YAML as a dictionary.

        Returns:
            The parsed stories or rules.
        """
        if not validate_training_data_format_version(parsed_content, self.source_name):
            return []

        for key, parser_class in {
            KEY_STORIES: StoryParser,
            KEY_RULES: RuleParser,
        }.items():
            data = parsed_content.get(key) or []
            parser = parser_class.from_reader(self)
            parser.parse_data(data)
            self.story_steps.extend(parser.get_steps())

        return self.story_steps

    @classmethod
    def is_stories_file(cls, file_path: Union[Text, Path]) -> bool:
        """Check if file contains Core training data or rule data in YAML format.

        Args:
            file_path: Path of the file to check.

        Returns:
            `True` in case the file is a Core YAML training data or rule data file,
            `False` otherwise.

        Raises:
            YamlException: if the file seems to be a YAML file (extension) but
                can not be read / parsed.
        """
        return rasa.shared.data.is_likely_yaml_file(file_path) and is_key_in_yaml(
            file_path, KEY_STORIES, KEY_RULES
        )

    @classmethod
    def _has_test_prefix(cls, file_path: Text) -> bool:
        """Check if the filename of a file at a path has a certain prefix.

        Arguments:
            file_path: path to the file

        Returns:
            `True` if the filename starts with the prefix, `False` otherwise.
        """
        return Path(file_path).name.startswith(TEST_STORIES_FILE_PREFIX)

    @classmethod
    def is_test_stories_file(cls, file_path: Union[Text, Path]) -> bool:
        """Checks if a file is a test conversations file.

        Args:
            file_path: Path of the file which should be checked.

        Returns:
            `True` if it's a conversation test file, otherwise `False`.
        """
        return cls._has_test_prefix(file_path) and cls.is_stories_file(file_path)

    def get_steps(self) -> List[StoryStep]:
        self._add_current_stories_to_result()
        return self.story_steps

    def parse_data(self, data: List[Dict[Text, Any]]) -> None:
        item_title = self._get_item_title()

        for item in data:
            if not isinstance(item, dict):
                rasa.shared.utils.io.raise_warning(
                    f"Unexpected block found in '{self.source_name}':\n"
                    f"{item}\nItems under the "
                    f"'{self._get_plural_item_title()}' key must be YAML "
                    f"dictionaries. It will be skipped.",
                    docs=self._get_docs_link(),
                )
                continue

            if item_title in item.keys():
                self._parse_plain_item(item)

    def _parse_plain_item(self, item: Dict[Text, Any]) -> None:
        item_name = item.get(self._get_item_title(), "")

        if not item_name:
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}': \n"
                f"{item}\n"
                f"The {self._get_item_title()} has an empty name. "
                f"{self._get_plural_item_title().capitalize()} should "
                f"have a name defined under '{self._get_item_title()}' "
                f"key. It will be skipped.",
                docs=self._get_docs_link(),
            )

        steps: List[Union[Text, Dict[Text, Any]]] = item.get(KEY_STEPS, [])

        if not steps:
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}': "
                f"The {self._get_item_title()} has no steps. "
                f"It will be skipped.",
                docs=self._get_docs_link(),
            )
            return

        self._new_part(item_name, item)

        for step in steps:
            self._parse_step(step)

        self._close_part(item)

    def _new_part(self, item_name: Text, item: Dict[Text, Any]) -> None:
        raise NotImplementedError()

    def _close_part(self, item: Dict[Text, Any]) -> None:
        pass

    def _parse_step(self, step: Union[Text, Dict[Text, Any]]) -> None:
        if isinstance(step, str):
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}':\n"
                f"Found an unexpected step in the {self._get_item_title()} "
                f"description:\n{step}\nThe step is of type `str` "
                f"which is only allowed for the rule snippet action "
                f"'{RULE_SNIPPET_ACTION_NAME}'. It will be skipped.",
                docs=self._get_docs_link(),
            )
        elif KEY_USER_INTENT in step.keys() or KEY_USER_MESSAGE in step.keys():
            self._parse_user_utterance(step)
        elif KEY_OR in step.keys():
            self._parse_or_statement(step)
        elif KEY_ACTION in step.keys():
            self._parse_action(step)
        elif KEY_BOT_END_TO_END_MESSAGE in step.keys():
            self._parse_bot_message(step)
        elif KEY_CHECKPOINT in step.keys():
            self._parse_checkpoint(step)
        # This has to be after the checkpoint test as there can be a slot key within
        # a checkpoint.
        elif KEY_SLOT_NAME in step.keys():
            self._parse_slot(step)
        elif KEY_ACTIVE_LOOP in step.keys():
            self._parse_active_loop(step[KEY_ACTIVE_LOOP])
        elif KEY_STACK_UPDATE in step.keys():
            self._parse_stack(step)
        elif KEY_METADATA in step.keys():
            pass
        else:
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}':\n"
                f"Found an unexpected step in the {self._get_item_title()} "
                f"description:\n{step}\nIt will be skipped.",
                docs=self._get_docs_link(),
            )

    def _get_item_title(self) -> Text:
        raise NotImplementedError()

    def _get_plural_item_title(self) -> Text:
        raise NotImplementedError()

    def _get_docs_link(self) -> Text:
        raise NotImplementedError()

    def _parse_user_utterance(self, step: Dict[Text, Any]) -> None:
        utterance = self._parse_raw_user_utterance(step)

        if not utterance:
            return

        is_end_to_end_utterance = KEY_USER_INTENT not in step
        if is_end_to_end_utterance:
            utterance.intent = {INTENT_NAME_KEY: None}
        else:
            self._validate_that_utterance_is_in_domain(utterance)

        if self.current_step_builder is not None:
            self.current_step_builder.add_user_messages([utterance])

    def _validate_that_utterance_is_in_domain(self, utterance: UserUttered) -> None:
        intent_name = utterance.intent.get(INTENT_NAME_KEY)

        # check if this is a retrieval intent
        # in this case check only for the base intent in domain
        intent_name = Message.separate_intent_response_key(intent_name)[0]

        if not self.domain:
            logger.debug(
                "Skipped validating if intent is in domain as domain " "is `None`."
            )
            return

        if intent_name not in self.domain.intents:
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}': \n"
                f"Found intent '{intent_name}' in stories which is not part of the "
                f"domain.",
                docs=DOCS_URL_STORIES,
            )

    def _parse_or_statement(self, step: Dict[Text, Any]) -> None:
        events: List = []

        for item in step.get(KEY_OR, []):
            if KEY_USER_INTENT in item.keys():
                utterance = self._parse_raw_user_utterance(item)
                if utterance:
                    events.append(utterance)
            elif KEY_CHECKPOINT_SLOTS in item.keys():
                for slot in item.get(KEY_CHECKPOINT_SLOTS, []):
                    if isinstance(slot, dict):
                        for key, value in slot.items():
                            parsed_events = self._parse_events(
                                SlotSet.type_name, {key: value}
                            )
                            events.extend(parsed_events)
                    elif isinstance(slot, str):
                        parsed_events = self._parse_events(
                            SlotSet.type_name, {slot: self._slot_default_value(slot)}
                        )
                        events.extend(parsed_events)
                    else:
                        rasa.shared.utils.io.raise_warning(
                            f"Issue found in '{self.source_name}':\n"
                            f"Invalid slot: \n{slot}\n"
                            f"Items under the '{KEY_CHECKPOINT_SLOTS}' key must be "
                            f"YAML dictionaries or Strings. "
                            f"The checkpoint will be skipped.",
                            docs=self._get_docs_link(),
                        )
                        return
            else:
                rasa.shared.utils.io.raise_warning(
                    f"Issue found in '{self.source_name}': \n"
                    f"`OR` statement can have '{KEY_USER_INTENT}' or '{KEY_SLOT_NAME}'"
                    f"as a sub-element. This step will be skipped:\n"
                    f"'{item}'\n",
                    docs=self._get_docs_link(),
                )
                return

        if events and self.current_step_builder is not None:
            self.current_step_builder.add_events(events)

    def _user_intent_from_step(
        self, step: Dict[Text, Any]
    ) -> Tuple[Text, Optional[Text]]:
        try:
            user_intent = step.get(KEY_USER_INTENT, "").strip()
        except AttributeError:
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}':\n"
                f"Missing intent value in {self._get_item_title()} step: {step} .",
                docs=self._get_docs_link(),
            )
            user_intent = ""

        if not user_intent and KEY_USER_MESSAGE not in step:
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}':\n"
                f"User utterance cannot be empty. "
                f"This {self._get_item_title()} step will be skipped:\n"
                f"{step}",
                docs=self._get_docs_link(),
            )

        if user_intent.startswith(INTENT_MESSAGE_PREFIX):
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}':\n"
                f"User intent '{user_intent}' starts with "
                f"'{INTENT_MESSAGE_PREFIX}'. This is not required.",
                docs=self._get_docs_link(),
            )
            # Remove leading slash
            user_intent = user_intent[1:]

        # StoryStep should never contain a full retrieval intent, only the base intent.
        # However, users can specify full retrieval intents in their test stories file
        # for the NLU testing purposes.
        base_intent, response_key = Message.separate_intent_response_key(user_intent)
        if response_key and not self.is_test_stories_file(self.source_name):
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}' while parsing story "
                f"{self._get_item_title()}:\n"
                f"User intent '{user_intent}' is a full retrieval intent. "
                f"Stories shouldn't contain full retrieval intents. "
                f"Rasa Pro will only use base intent '{base_intent}' "
                f"for training.",
                docs=self._get_docs_link(),
            )

        return (base_intent, user_intent) if response_key else (base_intent, None)

    def _parse_raw_user_utterance(self, step: Dict[Text, Any]) -> Optional[UserUttered]:
        intent_name, full_retrieval_intent = self._user_intent_from_step(step)
        intent = {
            INTENT_NAME_KEY: intent_name,
            FULL_RETRIEVAL_INTENT_NAME_KEY: full_retrieval_intent,
            PREDICTED_CONFIDENCE_KEY: 1.0,
        }

        if KEY_USER_MESSAGE in step:
            user_message = step[KEY_USER_MESSAGE].strip()
            entities = entities_parser.find_entities_in_training_example(user_message)
            plain_text = entities_parser.replace_entities(user_message)

            if plain_text.startswith(INTENT_MESSAGE_PREFIX):
                entities = self.unpack_regex_message(Message({TEXT: plain_text})).get(
                    ENTITIES, []
                )
        else:
            raw_entities = step.get(KEY_ENTITIES, [])
            entities = self._parse_raw_entities(raw_entities)
            # set plain_text to None because only intent was provided in the stories
            plain_text = None
        return UserUttered(
            plain_text,
            intent,
            entities,
            parse_data={KEY_COMMANDS: step.get(KEY_COMMANDS, [])},  # type: ignore[misc]
        )

    @staticmethod
    def _parse_raw_entities(
        raw_entities: Union[List[Dict[Text, Text]], List[Text]],
    ) -> List[Dict[Text, Optional[Text]]]:
        final_entities = []
        for entity in raw_entities:
            if isinstance(entity, dict):
                _entity_type = None
                _entity_value = None
                _entity_role = None
                _entity_group = None
                for key, value in entity.items():
                    if key == "role":
                        _entity_role = value
                    elif key == "group":
                        _entity_group = value
                    else:
                        _entity_type = key
                        _entity_value = value

                _entity_dict = {
                    "entity": _entity_type,
                    "value": _entity_value,
                    "role": _entity_role,
                    "group": _entity_group,
                }
                _entity_dict = {k: v for k, v in _entity_dict.items() if v is not None}
                final_entities.append(_entity_dict)
            else:
                final_entities.append({"entity": entity, "value": ""})

        return final_entities

    def _parse_slot(self, step: Dict[Text, Any]) -> None:
        for slot in step.get(KEY_CHECKPOINT_SLOTS, []):
            if isinstance(slot, dict):
                for key, value in slot.items():
                    self._add_event(SlotSet.type_name, {key: value})
            elif isinstance(slot, str):
                self._add_event(
                    SlotSet.type_name, {slot: self._slot_default_value(slot)}
                )
            else:
                rasa.shared.utils.io.raise_warning(
                    f"Issue found in '{self.source_name}':\n"
                    f"Invalid slot: \n{slot}\n"
                    f"Items under the '{KEY_CHECKPOINT_SLOTS}' key must be "
                    f"YAML dictionaries or Strings. The checkpoint will be skipped.",
                    docs=self._get_docs_link(),
                )
                return

    @functools.lru_cache()
    def _slot_default_value(self, slot_name: Text) -> Any:
        if not self.domain:
            return None

        slot_types_with_default_types = {
            TextSlot: DEFAULT_VALUE_TEXT_SLOTS,
            ListSlot: DEFAULT_VALUE_LIST_SLOTS,
        }
        slot = next(slot for slot in self.domain.slots if slot.name == slot_name)

        default_value = slot_types_with_default_types.get(type(slot))
        if default_value is None and slot.has_features():
            rasa.shared.utils.io.raise_warning(
                f"Slot '{slot_name}' was referenced by its name only. As slot "
                f"'{slot_name}' is of type '{slot.type_name}' you need to specify a "
                f"value for it. Slot '{slot_name}' will be treated as if it's value "
                f"is empty.",
                docs=DOCS_URL_SLOTS,
            )

        return default_value

    def _parse_action(self, step: Dict[Text, Any]) -> None:
        action_name = step.get(KEY_ACTION, "")
        if not action_name:
            rasa.shared.utils.io.raise_warning(
                f"Issue found in '{self.source_name}': \n"
                f"Action name cannot be empty. "
                f"This {self._get_item_title()} step will be skipped:\n"
                f"{step}",
                docs=self._get_docs_link(),
            )
            return

        self._add_event(action_name, {})

    def _parse_stack(self, step: Dict[Text, Any]) -> None:
        update = step.get(KEY_STACK_UPDATE, [])
        self._add_event(DialogueStackUpdated.type_name, {"update": json.dumps(update)})

    def _parse_bot_message(self, step: Dict[Text, Any]) -> None:
        bot_message = step.get(KEY_BOT_END_TO_END_MESSAGE, "")
        self._add_event("", {ACTION_TEXT: bot_message})

    def _parse_active_loop(self, active_loop_name: Optional[Text]) -> None:
        self._add_event(ActiveLoop.type_name, {LOOP_NAME: active_loop_name})

    def _parse_checkpoint(self, step: Dict[Text, Any]) -> None:
        checkpoint_name = step.get(KEY_CHECKPOINT, "")
        slots = step.get(KEY_CHECKPOINT_SLOTS, [])

        slots_dict = {}

        for slot in slots:
            if not isinstance(slot, dict):
                rasa.shared.utils.io.raise_warning(
                    f"Issue found in '{self.source_name}':\n"
                    f"Checkpoint '{checkpoint_name}' has an invalid slot: "
                    f"{slots}\nItems under the '{KEY_CHECKPOINT_SLOTS}' key must be "
                    f"YAML dictionaries. The checkpoint will be skipped.",
                    docs=self._get_docs_link(),
                )
                return

            for key, value in slot.items():
                slots_dict[key] = value

        self._add_checkpoint(checkpoint_name, slots_dict)

    @staticmethod
    def _regex_message_pattern() -> Pattern:
        """Builds the pattern that matches `TEXT`s of messages that need to be unpacked.

        Returns:
            pattern with named groups
        """
        return re.compile(INTENT_ENTITIES_PATTERN)

    @staticmethod
    def unpack_regex_message(
        message: Message,
        domain: Optional[Domain] = None,
        entity_extractor_name: Optional[Text] = None,
    ) -> Message:
        """Unpacks the message if `TEXT` contains an encoding of attributes.

        Args:
            message: some message
            domain: the domain
            entity_extractor_name: An extractor name which should be added for the
                entities.

        Returns:
            the given message if that message does not need to be unpacked, and a new
            message with the extracted attributes otherwise
        """
        user_text = message.get(TEXT).strip()

        # If the prefix doesn't match, we don't even need to try to match the pattern.
        if not user_text.startswith(INTENT_MESSAGE_PREFIX):
            return message

        # Try to match the pattern.
        match = YAMLStoryReader._regex_message_pattern().match(user_text)

        # If it doesn't match, then (potentially) something went wrong, because the
        # message text did start with the special prefix -- however, a user might
        # just have decided to start their text this way.
        if not match:
            structlogger.warning("message.parsing.failed")
            return message

        # Extract attributes from the match - and validate it via the domain.
        intent_name = YAMLStoryReader._intent_name_from_regex_match(match, domain)
        confidence = YAMLStoryReader._confidences_from_regex_match(match)
        entities = YAMLStoryReader._entities_from_regex_match(
            match, domain, entity_extractor_name
        )

        # The intent name is *not* optional, but during parsing we might find out
        # that the given intent is unknown (and warn). In this case, stop here.
        if intent_name is None:
            return message

        if match.group("rest"):
            rasa.shared.utils.io.raise_warning(
                f"Failed to parse arguments in line '{match.string}'. "
                f"Failed to interpret some parts. "
                f"Make sure your regex string is in the following format:"
                f"{INTENT_MESSAGE_PREFIX}"
                f"<intent_name>@<confidence-value><dictionary of entities> "
                f"Continuing without {match.group('rest')}. "
            )

        # Add the results to the message.
        intent_data = {
            INTENT_NAME_KEY: intent_name,
            PREDICTED_CONFIDENCE_KEY: confidence,
        }
        intent_ranking = [
            {INTENT_NAME_KEY: intent_name, PREDICTED_CONFIDENCE_KEY: confidence}
        ]
        message_data = {}
        message_data[TEXT] = user_text
        message_data[INTENT] = intent_data
        message_data[INTENT_RANKING_KEY] = intent_ranking
        message_data[ENTITIES] = entities
        return Message(message_data, output_properties=set(message_data.keys()))

    @staticmethod
    def _intent_name_from_regex_match(match: Match, domain: Domain) -> Optional[Text]:
        intent_name = match.group(INTENT_NAME_KEY).strip()
        if domain and intent_name not in domain.intents:
            rasa.shared.utils.io.raise_warning(
                f"Failed to parse arguments in line '{match.string}'. "
                f"Expected the intent to be one of [{domain.intents}] "
                f"but found {intent_name}."
                f"Continuing with given line as user text.",
                docs=DOCS_URL_STORIES,
            )
            intent_name = None
        return intent_name

    @staticmethod
    def _entities_from_regex_match(
        match: Match, domain: Domain, extractor_name: Optional[Text]
    ) -> List[Dict[Text, Any]]:
        """Extracts the optional entity information from the given pattern match.

        If no entities are specified or if the extraction fails, then an empty list
        is returned.

        Args:
            match: a match produced by `self.pattern`
            domain: the domain
            extractor_name: A extractor name which should be added for the entities

        Returns:
            some list of entities
        """
        entities_str = match.group(ENTITIES)
        if entities_str is None:
            return []

        try:
            parsed_entities = json.loads(entities_str)
            if not isinstance(parsed_entities, dict):
                raise ValueError(
                    f"Parsed value isn't a json object "
                    f"(instead parser found '{type(parsed_entities)}')"
                )
        except (JSONDecodeError, ValueError) as e:
            rasa.shared.utils.io.raise_warning(
                f"Failed to parse arguments in line '{match.string}'. "
                f"Failed to decode parameters as a json object (dict). "
                f"Make sure the intent is followed by a proper json object (dict). "
                f"Continuing without entities. "
                f"Error: {e}",
                docs=DOCS_URL_STORIES,
            )
            parsed_entities = dict()

        # validate the given entity types
        if domain:
            entity_types = set(parsed_entities.keys())
            unknown_entity_types = entity_types.difference(domain.entities)
            if unknown_entity_types:
                rasa.shared.utils.io.raise_warning(
                    f"Failed to parse arguments in line '{match.string}'. "
                    f"Expected entities from {domain.entities} "
                    f"but found {unknown_entity_types}. "
                    f"Continuing without unknown entity types. ",
                    docs=DOCS_URL_STORIES,
                )
                parsed_entities = {
                    key: value
                    for key, value in parsed_entities.items()
                    if key not in unknown_entity_types
                }

        # convert them into the list of dictionaries that we expect
        entities: List[Dict[Text, Any]] = []
        default_properties = {}
        if extractor_name:
            default_properties = {EXTRACTOR: extractor_name}

        for entity_type, entity_values in parsed_entities.items():
            if not isinstance(entity_values, list):
                entity_values = [entity_values]

            for entity_value in entity_values:
                entities.append(
                    {
                        ENTITY_ATTRIBUTE_TYPE: entity_type,
                        ENTITY_ATTRIBUTE_VALUE: entity_value,
                        ENTITY_ATTRIBUTE_START: match.start(ENTITIES),
                        ENTITY_ATTRIBUTE_END: match.end(ENTITIES),
                        **default_properties,
                    }
                )
        return entities

    @staticmethod
    def _confidences_from_regex_match(match: Match) -> float:
        """Extracts the optional confidence information from the given pattern match.

        If no confidence is specified, then this method returns the maximum
        confidence `1.0`.
        If a confidence is specified but extraction fails, then this method defaults
        to a confidence of `0.0`.

        Args:
            match: a match produced by `self.pattern`
            domain: the domain

        Returns:
            some confidence value
        """
        confidence_str = match.group(PREDICTED_CONFIDENCE_KEY)
        if confidence_str is None:
            return 1.0
        try:
            confidence_str = confidence_str.strip()[1:]  # remove the "@"
            try:
                confidence = float(confidence_str)
            except ValueError:
                confidence = 0.0
                raise ValueError(
                    f"Expected confidence to be a non-negative decimal number but "
                    f"found {confidence}. Continuing with 0.0 instead."
                )
            if confidence > 1.0:
                # Due to the pattern we know that this cannot be a negative number.
                original_confidence = confidence
                confidence = min(1.0, confidence)
                raise ValueError(
                    f"Expected confidence to be at most 1.0. "
                    f"but found {original_confidence}. "
                    f"Continuing with {confidence} instead."
                )
            return confidence

        except ValueError as e:
            rasa.shared.utils.io.raise_warning(
                f"Failed to parse arguments in line '{match.string}'. "
                f"Could not extract confidence value from `{confidence_str}'. "
                f"Make sure the intent confidence is an @ followed "
                f"by a decimal number that not negative and at most 1.0. "
                f"Error: {e}",
                docs=DOCS_URL_STORIES,
            )
            return confidence


class StoryParser(YAMLStoryReader):
    """Encapsulate story-specific parser behavior."""

    def _new_part(self, item_name: Text, item: Dict[Text, Any]) -> None:
        self._new_story_part(item_name, self.source_name)

    def _get_item_title(self) -> Text:
        return KEY_STORY_NAME

    def _get_plural_item_title(self) -> Text:
        return KEY_STORIES

    def _get_docs_link(self) -> Text:
        return DOCS_URL_STORIES


class RuleParser(YAMLStoryReader):
    """Encapsulate rule-specific parser behavior."""

    def _new_part(self, item_name: Text, item: Dict[Text, Any]) -> None:
        self._new_rule_part(item_name, self.source_name)
        conditions = item.get(KEY_RULE_CONDITION, [])
        self._parse_rule_conditions(conditions)
        if not item.get(KEY_RULE_FOR_CONVERSATION_START):
            self._parse_rule_snippet_action()

    def _parse_rule_conditions(
        self, conditions: List[Union[Text, Dict[Text, Any]]]
    ) -> None:
        self._is_parsing_conditions = True
        for condition in conditions:
            self._parse_step(condition)
        self._is_parsing_conditions = False

    def _close_part(self, item: Dict[Text, Any]) -> None:
        if item.get(KEY_WAIT_FOR_USER_INPUT_AFTER_RULE) is False:
            self._parse_rule_snippet_action()

    def _get_item_title(self) -> Text:
        return KEY_RULE_NAME

    def _get_plural_item_title(self) -> Text:
        return KEY_RULES

    def _get_docs_link(self) -> Text:
        return DOCS_URL_RULES

    def _parse_rule_snippet_action(self) -> None:
        self._add_event(RULE_SNIPPET_ACTION_NAME, {})



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Generator, Optional, Tuple

import structlog
from pydantic import BaseModel, Field

from rasa.e2e_test.constants import (
    DEFAULT_E2E_TESTING_MODEL,
    E2E_CONFIG_SCHEMA_FILE_PATH,
    KEY_EXTRA_PARAMETERS,
    KEY_LLM_E2E_TEST_CONVERSION,
    KEY_LLM_JUDGE,
)
from rasa.shared.constants import (
    API_BASE_CONFIG_KEY,
    DEPLOYMENT_CONFIG_KEY,
    EMBEDDINGS_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODELS_CONFIG_KEY,
    OPENAI_PROVIDER,
    PROVIDER_CONFIG_KEY,
)
from rasa.shared.utils.llm import (
    combine_custom_and_default_config,
    resolve_model_client_config,
)
from rasa.shared.utils.yaml import (
    parse_raw_yaml,
    read_schema_file,
    validate_yaml_content_using_schema,
)

structlogger = structlog.get_logger()

CONFTEST_PATTERNS = ["conftest.yml", "conftest.yaml"]


class BaseModelConfig(BaseModel):
    """Base class for model configurations used by generative assertions."""

    provider: Optional[str] = None
    model: Optional[str] = None
    extra_parameters: Dict[str, Any] = Field(default_factory=dict)
    model_group: Optional[str] = None


class LLMJudgeConfig(BaseModel):
    """Class for storing the configuration of the LLM-Judge.

    The LLM-Judge is used to measure the factual correctness
    (i.e., how grounded in the source documents the response is),
     or relevance of the generated response during E2E testing.
    """

    llm_config: BaseModelConfig
    embeddings: Optional[BaseModelConfig] = None

    @classmethod
    def get_default_llm_config(cls) -> Dict[str, Any]:
        return {
            PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
            MODEL_CONFIG_KEY: DEFAULT_E2E_TESTING_MODEL,
        }

    @classmethod
    def from_dict(cls, config_data: Dict[str, Any]) -> LLMJudgeConfig:
        """Loads the configuration from a dictionary."""
        embeddings = config_data.pop(EMBEDDINGS_CONFIG_KEY, {})
        llm_config = config_data.pop("llm", {})

        llm_config = resolve_model_client_config(llm_config)
        llm_config, llm_extra_parameters = cls.extract_attributes(llm_config)
        if not llm_config:
            llm_config = combine_custom_and_default_config(
                llm_config, cls.get_default_llm_config()
            )
        embeddings_config = resolve_model_client_config(embeddings)
        embeddings_config, embeddings_extra_parameters = cls.extract_attributes(
            embeddings_config
        )

        return LLMJudgeConfig(
            llm_config=BaseModelConfig(
                extra_parameters=llm_extra_parameters, **llm_config
            ),
            embeddings=BaseModelConfig(
                extra_parameters=embeddings_extra_parameters, **embeddings_config
            )
            if embeddings_config
            else None,
        )

    @classmethod
    def extract_attributes(
        cls, config: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Extract the expected fields from the configuration."""
        required_config = {}

        expected_fields = [
            PROVIDER_CONFIG_KEY,
            MODEL_CONFIG_KEY,
        ]

        if PROVIDER_CONFIG_KEY in config:
            required_config = {
                expected_field: config.pop(expected_field, None)
                for expected_field in expected_fields
            }

        elif MODELS_CONFIG_KEY in config:
            config = config.pop(MODELS_CONFIG_KEY)[0]

            required_config = {
                expected_field: config.pop(expected_field, None)
                for expected_field in expected_fields
            }

        clean_config = clean_up_config(required_config)
        return clean_config, config

    @property
    def llm_config_as_dict(self) -> Dict[str, Any]:
        return extract_config(self.llm_config)

    @property
    def embeddings_config_as_dict(self) -> Dict[str, Any]:
        if self.embeddings is None:
            return {}

        return extract_config(self.embeddings)


def clean_up_config(config_data: Dict[str, Any]) -> Dict[str, Any]:
    """Remove None values from the configuration."""
    return {key: value for key, value in config_data.items() if value}


def extract_config(config: BaseModelConfig) -> Dict[str, Any]:
    clean_config = clean_up_config(dict(config))
    extra_parameters = clean_config.pop(KEY_EXTRA_PARAMETERS, {})
    return {**clean_config, **extra_parameters}


class LLME2ETestConverterConfig(BaseModel):
    """Class for storing the LLM configuration of the E2ETestConverter.

    This configuration is used to initialize the LiteLLM client.
    """

    provider: Optional[str]
    model: Optional[str]
    deployment: Optional[str]
    api_base: Optional[str]
    extra_parameters: Optional[Dict[str, Any]]

    @classmethod
    def from_dict(cls, config_data: Dict[str, Any]) -> LLME2ETestConverterConfig:
        """Loads the configuration from a dictionary."""
        expected_fields = [
            PROVIDER_CONFIG_KEY,
            API_BASE_CONFIG_KEY,
            DEPLOYMENT_CONFIG_KEY,
            MODEL_CONFIG_KEY,
        ]
        kwargs = {
            expected_field: config_data.pop(expected_field, None)
            for expected_field in expected_fields
        }
        return cls(extra_parameters=config_data, **kwargs)

    @classmethod
    def get_default_config(cls) -> Dict[str, Any]:
        return {
            PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
            MODEL_CONFIG_KEY: DEFAULT_E2E_TESTING_MODEL,
        }

    @staticmethod
    def _clean_up_config(config_data: Dict[str, Any]) -> Dict[str, Any]:
        """Remove None values from the configuration."""
        return {key: value for key, value in config_data.items() if value}

    def as_dict(self) -> Dict[str, Any]:
        return self._clean_up_config(dict(self))


def get_conftest_path(test_case_path: Optional[Path]) -> Optional[Path]:
    """Get the path to the conftest.yml file.

    This assumes that the conftest.yml file is in the assistant project.
    """
    if test_case_path is None:
        return None

    while True:
        if test_case_path.is_file():
            test_case_path = test_case_path.parent

        matches = find_conftest_path(test_case_path)
        try:
            match = next(matches)
            structlogger.debug("e2e_config.get_conftest_path.found", match=match)
            return match
        except StopIteration:
            pass

        plausible_config_paths = [
            test_case_path / "config.yml",
            test_case_path / "config",
        ]
        for plausible_config_path in plausible_config_paths:
            if plausible_config_path.exists():
                # we reached the root of the assistant project
                return None
        # In case of an invalid path outside the assistant project,
        # break the loop if we reach the root
        if test_case_path == Path("."):
            return None

        test_case_path = test_case_path.parent


def find_conftest_path(path: Path) -> Generator[Path, None, None]:
    """Find the path to the conftest.yml file."""
    for pattern in CONFTEST_PATTERNS:
        for file_path in path.rglob(pattern):
            yield file_path


def create_llm_judge_config(test_case_path: Optional[Path]) -> LLMJudgeConfig:
    """Create the LLM-Judge configuration from the dictionary."""
    config_data = read_conftest_file(test_case_path)
    if not config_data:
        structlogger.debug("e2e_config.create_llm_judge_config.no_conftest_detected")
        return LLMJudgeConfig.from_dict(config_data)

    llm_judge_config_data = config_data.get(KEY_LLM_JUDGE, {})
    if not llm_judge_config_data:
        structlogger.debug("e2e_config.create_llm_judge_config.no_llm_judge_key")

    structlogger.info(
        "e2e_config.create_llm_judge_config.success",
        llm_judge_config_data=llm_judge_config_data,
    )

    return LLMJudgeConfig.from_dict(llm_judge_config_data)


def create_llm_e2e_test_converter_config(
    config_path: Path,
) -> LLME2ETestConverterConfig:
    """Create the LLME2ETestConverterConfig configuration from the dictionary."""
    config_data = read_conftest_file(config_path)
    if not config_data:
        structlogger.debug(
            "e2e_config.create_llm_e2e_test_converter_config.no_conftest_detected"
        )
        return LLME2ETestConverterConfig.from_dict(config_data)

    llm_e2e_test_converter_config_data = config_data.get(
        KEY_LLM_E2E_TEST_CONVERSION, {}
    )
    if not llm_e2e_test_converter_config_data:
        structlogger.debug(
            "e2e_config.create_llm_e2e_test_converter_config.no_llm_e2e_test_converter_config_key"
        )

    structlogger.info(
        "e2e_config.create_llm_e2e_test_converter_config.success",
        llm_e2e_test_converter_config_data=llm_e2e_test_converter_config_data,
    )

    return LLME2ETestConverterConfig.from_dict(llm_e2e_test_converter_config_data)


def read_conftest_file(test_case_path: Optional[Path]) -> Dict[str, Any]:
    """Read the conftest.yml file."""
    conftest_path = get_conftest_path(test_case_path)
    if conftest_path is None:
        return {}

    e2e_config_schema = read_schema_file(E2E_CONFIG_SCHEMA_FILE_PATH)
    config_data = parse_raw_yaml(conftest_path.read_text())
    validate_yaml_content_using_schema(config_data, e2e_config_schema)

    return config_data



from __future__ import annotations

import json
from inspect import isawaitable, iscoroutinefunction
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Text,
    TypeVar,
    Union,
)

import structlog

import rasa.shared.utils.cli
import rasa.shared.utils.common
import rasa.shared.utils.io
import rasa.utils.json_utils
from rasa.core.brokers.broker import EventBroker
from rasa.plugin import plugin_manager
from rasa.shared.core.constants import ACTION_LISTEN_NAME
from rasa.shared.core.conversation import Dialogue
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event
from rasa.shared.core.trackers import (
    ActionExecuted,
    DialogueStateTracker,
    TrackerEventDiffEngine,
)
from rasa.shared.exceptions import ConnectionException, RasaException
from rasa.utils.endpoints import EndpointConfig

structlogger = structlog.get_logger(__name__)


def check_if_tracker_store_async(tracker_store: TrackerStore) -> bool:
    """Evaluates if a tracker store object is async based on implementation of methods.

    :param tracker_store: tracker store object we're evaluating
    :return: if the tracker store correctly implements all async methods
    """
    return all(
        iscoroutinefunction(getattr(tracker_store, method))
        for method in _get_async_tracker_store_methods()
    )


def _get_async_tracker_store_methods() -> List[str]:
    return [
        attribute
        for attribute in dir(TrackerStore)
        if iscoroutinefunction(getattr(TrackerStore, attribute))
    ]


class TrackerDeserialisationException(RasaException):
    """Raised when an error is encountered while deserialising a tracker."""


SerializationType = TypeVar("SerializationType")


class SerializedTrackerRepresentation(Generic[SerializationType]):
    """Mixin class for specifying different serialization methods per tracker store."""

    @staticmethod
    def serialise_tracker(tracker: DialogueStateTracker) -> SerializationType:
        """Requires implementation to return representation of tracker."""
        raise NotImplementedError()


class SerializedTrackerAsText(SerializedTrackerRepresentation[Text]):
    """Mixin class that returns the serialized tracker as string."""

    @staticmethod
    def serialise_tracker(tracker: DialogueStateTracker) -> Text:
        """Serializes the tracker, returns representation of the tracker."""
        dialogue = tracker.as_dialogue()

        return json.dumps(dialogue.as_dict())


class SerializedTrackerAsDict(SerializedTrackerRepresentation[Dict]):
    """Mixin class that returns the serialized tracker as dictionary."""

    @staticmethod
    def serialise_tracker(tracker: DialogueStateTracker) -> Dict:
        """Serializes the tracker, returns representation of the tracker."""
        d = tracker.as_dialogue().as_dict()
        d.update({"sender_id": tracker.sender_id})
        return d


class TrackerStore:
    """Represents common behavior and interface for all `TrackerStore`s."""

    def __init__(
        self,
        domain: Optional[Domain],
        event_broker: Optional[EventBroker] = None,
        **kwargs: Dict[Text, Any],
    ) -> None:
        """Create a TrackerStore.

        Args:
            domain: The `Domain` to initialize the `DialogueStateTracker`.
            event_broker: An event broker to publish any new events to another
                destination.
            kwargs: Additional kwargs.
        """
        self._domain = domain or Domain.empty()
        self.event_broker = event_broker
        self.max_event_history: Optional[int] = None

    @staticmethod
    def create(
        obj: Union[TrackerStore, EndpointConfig, None],
        domain: Optional[Domain] = None,
        event_broker: Optional[EventBroker] = None,
    ) -> TrackerStore:
        """Factory to create a tracker store."""
        if isinstance(obj, TrackerStore):
            return obj

        import pymongo.errors
        import sqlalchemy.exc
        from botocore.exceptions import BotoCoreError, ClientError

        try:
            _tracker_store = plugin_manager().hook.create_tracker_store(
                endpoint_config=obj,
                domain=domain,
                event_broker=event_broker,
            )

            tracker_store = (
                _tracker_store
                if _tracker_store
                else create_tracker_store(obj, domain, event_broker)
            )

            return tracker_store
        except (
            BotoCoreError,
            ClientError,
            pymongo.errors.ConnectionFailure,
            sqlalchemy.exc.OperationalError,
            ConnectionError,
            pymongo.errors.OperationFailure,
        ) as error:
            raise ConnectionException(
                "Cannot connect to tracker store." + str(error)
            ) from error

    async def get_or_create_tracker(
        self,
        sender_id: Text,
        max_event_history: Optional[int] = None,
        append_action_listen: bool = True,
    ) -> "DialogueStateTracker":
        """Returns tracker or creates one if the retrieval returns None.

        Args:
            sender_id: Conversation ID associated with the requested tracker.
            max_event_history: Value to update the tracker store's max event history to.
            append_action_listen: Whether or not to append an initial `action_listen`.
        """
        self.max_event_history = max_event_history

        tracker = await self.retrieve(sender_id)

        if tracker is None:
            tracker = await self.create_tracker(
                sender_id, append_action_listen=append_action_listen
            )

        return tracker

    def init_tracker(self, sender_id: Text) -> "DialogueStateTracker":
        """Returns a Dialogue State Tracker."""
        return DialogueStateTracker(
            sender_id,
            self.domain.slots,
            max_event_history=self.max_event_history,
        )

    async def create_tracker(
        self, sender_id: Text, append_action_listen: bool = True
    ) -> DialogueStateTracker:
        """Creates a new tracker for `sender_id`.

        The tracker begins with a `SessionStarted` event and is initially listening.

        Args:
            sender_id: Conversation ID associated with the tracker.
            append_action_listen: Whether or not to append an initial `action_listen`.

        Returns:
            The newly created tracker for `sender_id`.
        """
        tracker = self.init_tracker(sender_id)

        if append_action_listen:
            tracker.update(ActionExecuted(ACTION_LISTEN_NAME))

        await self.save(tracker)

        return tracker

    async def save(self, tracker: DialogueStateTracker) -> None:
        """Save method that will be overridden by specific tracker."""
        raise NotImplementedError()

    async def exists(self, conversation_id: Text) -> bool:
        """Checks if tracker exists for the specified ID.

        This method may be overridden by the specific tracker store for
        faster implementations.

        Args:
            conversation_id: Conversation ID to check if the tracker exists.

        Returns:
            `True` if the tracker exists, `False` otherwise.
        """
        return await self.retrieve(conversation_id) is not None

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Retrieves tracker for the latest conversation session.

        This method will be overridden by the specific tracker store.

        Args:
            sender_id: Conversation ID to fetch the tracker for.

        Returns:
            Tracker containing events from the latest conversation sessions.
        """
        raise NotImplementedError()

    async def delete(self, sender_id: str) -> None:
        """Delete tracker for the given sender_id.

        Args:
            sender_id: Conversation ID to delete the tracker for.
        """
        raise NotImplementedError()

    async def update(
        self, tracker: DialogueStateTracker, apply_deletion_only: bool = True
    ) -> None:
        """Replace an existing tracker with a new one.

        Args:
            tracker: The tracker to update.
            apply_deletion_only: If True (default), only a trim/delete step is applied
                (SQL: timestamp-based delete). If False, content-only updates may be
                applied when no rows are deleted (e.g. anonymization). Other stores
                ignore this; SQL uses it to avoid conflating deletion with content-only.
        """
        raise NotImplementedError()

    async def retrieve_full_tracker(
        self, conversation_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Retrieve method for fetching all tracker events.

        Fetches events across conversation sessions. The default implementation
        uses `self.retrieve()`.

        Args:
            conversation_id: The conversation ID to retrieve the tracker for.

        Returns:
            The fetch tracker containing all events across session starts.
        """
        return await self.retrieve(conversation_id)

    async def get_or_create_full_tracker(
        self,
        sender_id: Text,
        append_action_listen: bool = True,
    ) -> "DialogueStateTracker":
        """Returns tracker or creates one if the retrieval returns None.

        Args:
            sender_id: Conversation ID associated with the requested tracker.
            append_action_listen: Whether to append an initial `action_listen`.

        Returns:
            The tracker for the conversation ID.
        """
        tracker = await self.retrieve_full_tracker(sender_id)

        if tracker is None:
            tracker = await self.create_tracker(
                sender_id, append_action_listen=append_action_listen
            )

        return tracker

    async def stream_events(self, tracker: DialogueStateTracker) -> None:
        """Streams events to a message broker."""
        if self.event_broker is None:
            structlogger.debug(
                "tracker_store.stream_events.no_broker_configured",
                event_info="No event broker configured. Skipping streaming events.",
            )
            return None

        if (
            hasattr(self.event_broker, "stream_pii")
            and not self.event_broker.stream_pii
        ):
            # If the event broker is configured to not stream un-anonymized events,
            # skip streaming
            structlogger.debug(
                "tracker_store.stream_events.no_streaming",
                event_info="Un-anonymized events will not be published "
                "to the event broker.",
            )
            return None

        old_tracker = await self.retrieve(tracker.sender_id)
        new_events = TrackerEventDiffEngine.event_difference(old_tracker, tracker)

        await self._stream_new_events(self.event_broker, new_events, tracker.sender_id)

    async def _stream_new_events(
        self,
        event_broker: EventBroker,
        new_events: List[Event],
        sender_id: Text,
    ) -> None:
        """Publishes new tracker events to a message broker."""
        for event in new_events:
            body = {"sender_id": sender_id}
            body.update(event.as_dict())
            event_broker.publish(body)

    async def keys(self) -> Iterable[Text]:
        """Returns the set of values for the tracker store's primary key."""
        raise NotImplementedError()

    async def count_conversations(self, after_timestamp: float = 0.0) -> int:
        """Returns the number of conversations that have occurred after a timestamp.

        By default, this method returns the number of conversations that
        have occurred after the Unix epoch (i.e. timestamp 0). A conversation
        is considered to have occurred after a timestamp if at least one event
        happened after that timestamp.
        """
        tracker_keys = await self.keys()

        conversation_count = 0
        for key in tracker_keys:
            tracker = await self.retrieve(key)
            if tracker is None or not tracker.events:
                continue

            last_event = tracker.events[-1]
            if last_event.timestamp >= after_timestamp:
                conversation_count += 1

        return conversation_count

    def deserialise_tracker(
        self, sender_id: Text, serialised_tracker: Union[Text, bytes]
    ) -> Optional[DialogueStateTracker]:
        """Deserializes the tracker and returns it."""
        tracker = self.init_tracker(sender_id)

        try:
            dialogue = Dialogue.from_parameters(json.loads(serialised_tracker))
        except UnicodeDecodeError as e:
            raise TrackerDeserialisationException(
                "Tracker cannot be deserialised. "
                "Trackers must be serialised as json. "
                "Support for deserialising pickled trackers has been removed."
            ) from e

        tracker.recreate_from_dialogue(dialogue)

        return tracker

    @property
    def domain(self) -> Domain:
        """Returns the domain of the tracker store."""
        return self._domain

    @domain.setter
    def domain(self, domain: Optional[Domain]) -> None:
        self._domain = domain or Domain.empty()


class InMemoryTrackerStore(TrackerStore, SerializedTrackerAsText):
    """Stores conversation history in memory."""

    def __init__(
        self,
        domain: Domain,
        event_broker: Optional[EventBroker] = None,
        **kwargs: Dict[Text, Any],
    ) -> None:
        """Initializes the tracker store."""
        self.store: Dict[Text, Text] = {}
        super().__init__(domain, event_broker, **kwargs)

    async def save(self, tracker: DialogueStateTracker) -> None:
        """Updates and saves the current conversation state."""
        await self.stream_events(tracker)
        serialised = InMemoryTrackerStore.serialise_tracker(tracker)
        self.store[tracker.sender_id] = serialised

    async def delete(self, sender_id: Text) -> None:
        """Delete tracker for the given sender_id."""
        if sender_id not in self.store:
            structlogger.info(
                "in_memory_tracker_store.delete.no_tracker_for_sender_id",
                event_info=f"Could not find tracker for conversation ID '{sender_id}'.",
            )
            return None

        del self.store[sender_id]

        structlogger.info(
            "in_memory_tracker_store.delete.deleted_tracker",
            sender_id=sender_id,
        )

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Returns tracker matching sender_id."""
        return await self._retrieve(sender_id, fetch_all_sessions=False)

    async def keys(self) -> Iterable[Text]:
        """Returns sender_ids of the Tracker Store in memory."""
        return self.store.keys()

    async def retrieve_full_tracker(
        self, sender_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Returns tracker matching sender_id.

        Args:
            sender_id: Conversation ID to fetch the tracker for.
        """
        return await self._retrieve(sender_id, fetch_all_sessions=True)

    async def _retrieve(
        self, sender_id: Text, fetch_all_sessions: bool
    ) -> Optional[DialogueStateTracker]:
        """Returns tracker matching sender_id.

        Args:
            sender_id: Conversation ID to fetch the tracker for.
            fetch_all_sessions: Whether to fetch all sessions or only the last one.
        """
        if sender_id not in self.store:
            structlogger.debug(
                "in_memory_tracker_store.retrieve.no_tracker_for_sender_id",
                event_info=f"Could not find tracker for conversation ID '{sender_id}'.",
            )
            return None

        tracker = self.deserialise_tracker(sender_id, self.store[sender_id])

        if not tracker:
            structlogger.debug(
                "in_memory_tracker_store.retrieve.failed_to_deserialize_tracker",
                event_info=(
                    f"Could not deserialize tracker "
                    f"for conversation ID '{sender_id}'.",
                ),
            )
            return None

        if fetch_all_sessions:
            return tracker

        # only return the last session
        multiple_tracker_sessions = (
            rasa.shared.core.trackers.get_trackers_for_conversation_sessions(tracker)
        )

        if len(multiple_tracker_sessions) <= 1:
            return tracker

        return multiple_tracker_sessions[-1]

    async def update(
        self, tracker: DialogueStateTracker, apply_deletion_only: bool = True
    ) -> None:
        """Replace an existing tracker with a new one.

        Args:
            tracker: The tracker to update.
            apply_deletion_only: Ignored for in-memory store (full overwrite).
        """
        await self.save(tracker)


def validate_port(port: Any) -> Optional[int]:
    """Ensure that port can be converted to integer.

    Raises:
        RasaException if port cannot be cast to integer.
    """
    if port is not None and not isinstance(port, int):
        try:
            port = int(port)
        except ValueError as e:
            raise RasaException(f"The port '{port}' cannot be cast to integer.") from e

    return port


class FailSafeTrackerStore(TrackerStore):
    """Tracker store wrapper.

    Allows a fallback to a different tracker store in case of errors.
    """

    def __init__(
        self,
        tracker_store: TrackerStore,
        on_tracker_store_error: Optional[Callable[[Exception], None]] = None,
        fallback_tracker_store: Optional[TrackerStore] = None,
    ) -> None:
        """Create a `FailSafeTrackerStore`.

        Args:
            tracker_store: Primary tracker store.
            on_tracker_store_error: Callback which is called when there is an error
                in the primary tracker store.
            fallback_tracker_store: Fallback tracker store.
        """
        self._fallback_tracker_store: Optional[TrackerStore] = fallback_tracker_store
        self._tracker_store = tracker_store
        self._on_tracker_store_error = on_tracker_store_error

        super().__init__(tracker_store.domain, tracker_store.event_broker)

    @property
    def domain(self) -> Domain:
        """Returns the domain of the primary tracker store."""
        return self._tracker_store.domain

    @domain.setter
    def domain(self, domain: Optional[Domain]) -> None:
        self._tracker_store.domain = domain

        if self._fallback_tracker_store:
            self._fallback_tracker_store.domain = domain

    @property
    def fallback_tracker_store(self) -> TrackerStore:
        """Returns the fallback tracker store."""
        if not self._fallback_tracker_store:
            self._fallback_tracker_store = InMemoryTrackerStore(
                self._tracker_store.domain, self._tracker_store.event_broker
            )

        return self._fallback_tracker_store

    def on_tracker_store_error(self, error: Exception) -> None:
        """Calls the callback when there is an error in the primary tracker store."""
        if self._on_tracker_store_error:
            self._on_tracker_store_error(error)
        else:
            structlogger.error(
                "fail_safe_tracker_store.tracker_store_error",
                event_info=(
                    f"Error happened when trying to save conversation tracker to "
                    f"'{self._tracker_store.__class__.__name__}'. Falling back to use "
                    f"the '{InMemoryTrackerStore.__name__}'. Please "
                    f"investigate the following error: {error}."
                ),
            )

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Calls `retrieve` method of primary tracker store."""
        try:
            return await self._tracker_store.retrieve(sender_id)
        except Exception as e:
            self.on_tracker_store_retrieve_error(e)
            return None

    async def keys(self) -> Iterable[Text]:
        """Calls `keys` method of primary tracker store."""
        try:
            return await self._tracker_store.keys()
        except Exception as e:
            self.on_tracker_store_error(e)
            return []

    async def save(self, tracker: DialogueStateTracker) -> None:
        """Calls `save` method of primary tracker store."""
        try:
            await self._tracker_store.save(tracker)
        except Exception as e:
            self.on_tracker_store_error(e)
            await self.fallback_tracker_store.save(tracker)

    async def delete(self, sender_id: Text) -> None:
        """Delete tracker for the given sender_id."""
        try:
            await self._tracker_store.delete(sender_id)
        except Exception as e:
            self.on_tracker_store_error(e)
            await self.fallback_tracker_store.delete(sender_id)

    async def update(
        self, tracker: DialogueStateTracker, apply_deletion_only: bool = True
    ) -> None:
        """Replace an existing tracker with a new one."""
        try:
            await self._tracker_store.update(
                tracker, apply_deletion_only=apply_deletion_only
            )
        except Exception as e:
            self.on_tracker_store_error(e)
            await self.fallback_tracker_store.update(
                tracker, apply_deletion_only=apply_deletion_only
            )

    async def retrieve_full_tracker(
        self, sender_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Calls `retrieve_full_tracker` method of primary tracker store.

        Args:
            sender_id: The sender id of the tracker to retrieve.
        """
        try:
            return await self._tracker_store.retrieve_full_tracker(sender_id)
        except Exception as e:
            self.on_tracker_store_retrieve_error(e)
            return None

    def on_tracker_store_retrieve_error(self, error: Exception) -> None:
        """Calls `_on_tracker_store_error` callable attribute if set.

        Otherwise, logs the error.

        Args:
            error: The error that occurred.
        """
        if self._on_tracker_store_error:
            self._on_tracker_store_error(error)
        else:
            structlogger.error(
                "fail_safe_tracker_store.tracker_store_retrieve_error",
                event_info=(
                    f"Error happened when trying to retrieve conversation tracker from "
                    f"'{self._tracker_store.__class__.__name__}'. Falling back to use "
                    f"the '{InMemoryTrackerStore.__name__}'."
                ),
                exec_info=error,
            )


def _create_from_endpoint_config(
    endpoint_config: Optional[EndpointConfig] = None,
    domain: Optional[Domain] = None,
    event_broker: Optional[EventBroker] = None,
) -> TrackerStore:
    """Given an endpoint configuration, create a proper tracker store object."""
    from rasa.core.tracker_stores.dynamo_tracker_store import DynamoTrackerStore
    from rasa.core.tracker_stores.mongo_tracker_store import MongoTrackerStore
    from rasa.core.tracker_stores.redis_tracker_store import (
        RedisTrackerStore,
    )
    from rasa.core.tracker_stores.sql_tracker_store import SQLTrackerStore

    domain = domain or Domain.empty()

    if endpoint_config is None or endpoint_config.type is None:
        # default tracker store if no type is set
        tracker_store: TrackerStore = InMemoryTrackerStore(domain, event_broker)
    elif endpoint_config.type.lower() == "redis":
        tracker_store = RedisTrackerStore(
            domain=domain,
            host=endpoint_config.url,
            event_broker=event_broker,
            **endpoint_config.kwargs,
        )
    elif endpoint_config.type.lower() == "mongod":
        tracker_store = MongoTrackerStore(
            domain=domain,
            host=endpoint_config.url,
            event_broker=event_broker,
            **endpoint_config.kwargs,
        )
    elif endpoint_config.type.lower() == "sql":
        tracker_store = SQLTrackerStore(
            domain=domain,
            host=endpoint_config.url,
            event_broker=event_broker,
            **endpoint_config.kwargs,
        )
    elif endpoint_config.type.lower() == "dynamo":
        tracker_store = DynamoTrackerStore(
            domain=domain, event_broker=event_broker, **endpoint_config.kwargs
        )
    else:
        tracker_store = _load_from_module_name_in_endpoint_config(
            domain, endpoint_config, event_broker
        )

    structlogger.debug(
        "tracker_store.create_tracker_store_from_endpoint_config",
        event_info=f"Connected to {tracker_store.__class__.__name__}.",
    )

    return tracker_store


def _load_from_module_name_in_endpoint_config(
    domain: Domain, store: EndpointConfig, event_broker: Optional[EventBroker] = None
) -> TrackerStore:
    """Initializes a custom tracker.

    Defaults to the InMemoryTrackerStore if the module path can not be found.

    Args:
        domain: defines the universe in which the assistant operates
        store: the specific tracker store
        event_broker: an event broker to publish events

    Returns:
        a tracker store from a specified type in a stores endpoint configuration
    """
    try:
        tracker_store_class = rasa.shared.utils.common.class_from_module_path(
            store.type
        )

        return tracker_store_class(
            host=store.url, domain=domain, event_broker=event_broker, **store.kwargs
        )
    except (AttributeError, ImportError):
        rasa.shared.utils.io.raise_warning(
            f"Tracker store with type '{store.type}' not found. "
            f"Using `InMemoryTrackerStore` instead."
        )
        return InMemoryTrackerStore(domain)


def create_tracker_store(
    endpoint_config: Optional[EndpointConfig],
    domain: Optional[Domain] = None,
    event_broker: Optional[EventBroker] = None,
) -> TrackerStore:
    """Creates a tracker store based on the current configuration."""
    tracker_store = _create_from_endpoint_config(endpoint_config, domain, event_broker)

    if not check_if_tracker_store_async(tracker_store):
        rasa.shared.utils.io.raise_deprecation_warning(
            f"Tracker store implementation "
            f"{tracker_store.__class__.__name__} "
            f"is not asynchronous. Non-asynchronous tracker stores "
            f"are currently deprecated and will be removed in 4.0. "
            f"Please make the following methods async: "
            f"{_get_async_tracker_store_methods()}"
        )
        tracker_store = AwaitableTrackerStore(tracker_store)

    return tracker_store


class AwaitableTrackerStore(TrackerStore):
    """Wraps a tracker store so it can be implemented with async overrides."""

    def __init__(
        self,
        tracker_store: TrackerStore,
    ) -> None:
        """Create a `AwaitableTrackerStore`.

        Args:
            tracker_store: the wrapped tracker store.
        """
        self._tracker_store = tracker_store

        super().__init__(tracker_store.domain, tracker_store.event_broker)

    @property
    def domain(self) -> Domain:
        """Returns the domain of the primary tracker store."""
        return self._tracker_store.domain

    @domain.setter
    def domain(self, domain: Optional[Domain]) -> None:
        """Setter method to modify the wrapped tracker store's domain field."""
        self._tracker_store.domain = domain or Domain.empty()

    @staticmethod
    def create(
        obj: Union[TrackerStore, EndpointConfig, None],
        domain: Optional[Domain] = None,
        event_broker: Optional[EventBroker] = None,
    ) -> TrackerStore:
        """Wrapper to call `create` method of primary tracker store."""
        if isinstance(obj, TrackerStore):
            return AwaitableTrackerStore(obj)
        elif isinstance(obj, EndpointConfig):
            return AwaitableTrackerStore(_create_from_endpoint_config(obj))
        else:
            raise ValueError(
                f"{type(obj).__name__} supplied "
                f"but expected object of type {TrackerStore.__name__} or "
                f"of type {EndpointConfig.__name__}."
            )

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Wrapper to call `retrieve` method of primary tracker store."""
        result = self._tracker_store.retrieve(sender_id)
        return await result if isawaitable(result) else result

    async def keys(self) -> Iterable[Text]:
        """Wrapper to call `keys` method of primary tracker store."""
        result = self._tracker_store.keys()
        return await result if isawaitable(result) else result

    async def save(self, tracker: DialogueStateTracker) -> None:
        """Wrapper to call `save` method of primary tracker store."""
        result = self._tracker_store.save(tracker)
        return await result if isawaitable(result) else result

    async def delete(self, sender_id: Text) -> None:
        """Delete tracker for the given sender_id."""
        result = self._tracker_store.delete(sender_id)
        return await result if isawaitable(result) else result

    async def update(
        self, tracker: DialogueStateTracker, apply_deletion_only: bool = True
    ) -> None:
        """Replace an existing tracker with a new one."""
        result = self._tracker_store.update(
            tracker, apply_deletion_only=apply_deletion_only
        )
        return await result if isawaitable(result) else result

    async def retrieve_full_tracker(
        self, conversation_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Wrapper to call `retrieve_full_tracker` method of primary tracker store."""
        result = self._tracker_store.retrieve_full_tracker(conversation_id)
        return await result if isawaitable(result) else result



import logging
import math
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Text,
    Tuple,
    Union,
    cast,
)

import numpy as np
import scipy.sparse
import tensorflow as tf
from tensorflow.keras.utils import Sequence

from rasa.utils.tensorflow.constants import BALANCED, SEQUENCE
from rasa.utils.tensorflow.model_data import (
    Data,
    FeatureArray,
    FeatureSignature,
    RasaModelData,
)

logger = logging.getLogger(__name__)


class RasaDataGenerator(Sequence):
    """Abstract data generator."""

    def __init__(
        self,
        model_data: RasaModelData,
        batch_size: Union[int, List[int]],
        batch_strategy: Text = SEQUENCE,
        shuffle: bool = True,
    ):
        """Initializes the data generator.

        Args:
            model_data: The model data to use.
            batch_size: The batch size(s).
            batch_strategy: The batch strategy.
            shuffle: If 'True', data should be shuffled.
        """
        self.model_data = model_data
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.batch_strategy = batch_strategy

    def __len__(self) -> int:
        """Number of batches in the Sequence.

        Returns:
            The number of batches in the Sequence.
        """
        raise NotImplementedError

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """Gets batch at position `index`.

        Arguments:
            index: position of the batch in the Sequence.

        Returns:
            A batch (tuple of input data and target data).
        """
        raise NotImplementedError

    def on_epoch_end(self) -> None:
        """Update the data after every epoch."""
        raise NotImplementedError

    def _shuffle_and_balance(self, batch_size: int) -> Data:
        data = self.model_data.data

        if self.shuffle:
            data = self.model_data.shuffled_data(data)

        if self.batch_strategy == BALANCED:
            data = self.model_data.balanced_data(data, batch_size, self.shuffle)

        # do not override self.model_data.data, because we need original data for
        # balancing on the next epoch
        return data

    @staticmethod
    def _create_default_array() -> np.ndarray:
        """Create a default empty array for missing features.

        Returns:
            A default empty array with shape (0, 1) and dtype float32.
        """
        return np.zeros((0, 1), dtype=np.float32)

    @staticmethod
    def prepare_batch(
        data: Data,
        start: Optional[int] = None,
        end: Optional[int] = None,
        tuple_sizes: Optional[Dict[Text, int]] = None,
    ) -> Tuple[np.ndarray, ...]:
        """Slices model data into batch using given start and end value.

        Args:
            data: The data to prepare.
            start: The start index of the batch
            end: The end index of the batch
            tuple_sizes: In case the feature is not present we propagate the batch with
              default arrays. Tuple sizes contains the number of how many default values
              to add for what kind of feature.

        Returns:
            The features of the batch.
        """
        batch_data = []

        for key, attribute_data in data.items():
            for sub_key, f_data in attribute_data.items():
                # add default arrays for not present values during processing
                if not f_data:
                    if tuple_sizes:
                        batch_data += [
                            RasaDataGenerator._create_default_array()
                        ] * tuple_sizes[key]
                    else:
                        batch_data.append(RasaDataGenerator._create_default_array())
                    continue

                for v in f_data:
                    if start is not None and end is not None:
                        _data = v[start:end]
                    elif start is not None:
                        _data = v[start:]
                    elif end is not None:
                        _data = v[:end]
                    else:
                        _data = v[:]

                    if cast(FeatureArray, _data).is_sparse:
                        batch_data.extend(
                            RasaDataGenerator._scipy_matrix_to_values(_data)
                        )
                    else:
                        batch_data.append(RasaDataGenerator._pad_dense_data(_data))

        # len of batch_data is equal to the number of keys in model data
        return tuple(batch_data)

    @staticmethod
    def _pad_dense_data(array_of_dense: FeatureArray) -> np.ndarray:
        """Pad data of different lengths.

        Sequential data is padded with zeros. Zeros are added to the end of data.

        Args:
            array_of_dense: The array to pad.

        Returns:
            The padded array.
        """
        if array_of_dense.number_of_dimensions == 4:
            return RasaDataGenerator._pad_4d_dense_data(array_of_dense)

        if array_of_dense[0].ndim < 2:
            # data doesn't contain a sequence
            return array_of_dense.astype(np.float32)

        data_size = len(array_of_dense)
        max_seq_len = max([x.shape[0] for x in array_of_dense])

        data_padded = np.zeros(
            [data_size, max_seq_len, array_of_dense[0].shape[-1]],
            dtype=array_of_dense[0].dtype,
        )
        for i in range(data_size):
            data_padded[i, : array_of_dense[i].shape[0], :] = array_of_dense[i]

        return data_padded.astype(np.float32)

    @staticmethod
    def _pad_4d_dense_data(feature_array: FeatureArray) -> np.ndarray:
        # in case of dialogue data we may have 4 dimensions
        # batch size x dialogue history length x sequence length x number of features

        # as transformers cannot handle 4D tensors pad and reshape the data
        # so that the resulting tensor is 3D
        # the shape is (sum of dialogue history length for all tensors in the
        # batch x max sequence length x number of features)
        # the original shape and the original dialogue length is passed on to the model
        # it can be used to transform the 3D tensor back into 4D

        # in order to create 4d tensor inputs, we created "fake" zero features
        # for nonexistent inputs. To save calculation we filter this features before
        # input to tf methods.
        number_of_features = feature_array[0][0].shape[-1]
        array_of_array_of_dense = RasaDataGenerator._filter_out_fake_inputs(
            feature_array
        )
        if not array_of_array_of_dense:
            # return empty 3d array with appropriate last dims
            return np.zeros((0, 0, number_of_features), dtype=np.float32)

        combined_dialogue_len = sum(
            len(array_of_dense) for array_of_dense in array_of_array_of_dense
        )
        max_seq_len = max(
            [
                x.shape[0]
                for array_of_dense in array_of_array_of_dense
                for x in array_of_dense
            ]
        )

        data_padded = np.zeros(
            [combined_dialogue_len, max_seq_len, number_of_features],
            dtype=array_of_array_of_dense[0][0].dtype,
        )

        current_sum_dialogue_len = 0
        for i, array_of_dense in enumerate(array_of_array_of_dense):
            for j, dense in enumerate(array_of_dense):
                data_padded[current_sum_dialogue_len + j, : dense.shape[0], :] = dense
            current_sum_dialogue_len += len(array_of_dense)

        return data_padded.astype(np.float32)

    @staticmethod
    def _scipy_matrix_to_values(array_of_sparse: FeatureArray) -> List[np.ndarray]:
        """Convert a scipy matrix into indices, data, and shape.

        Args:
            array_of_sparse: The sparse data array.

        Returns:
            A list of dense numpy arrays representing the sparse data.
        """
        if array_of_sparse.number_of_dimensions == 4:
            return RasaDataGenerator._4d_scipy_matrix_to_values(array_of_sparse)

        # we need to make sure that the matrices are coo_matrices otherwise the
        # transformation does not work (e.g. you cannot access x.row, x.col)
        if not isinstance(array_of_sparse[0], scipy.sparse.coo_matrix):
            array_of_sparse = [x.tocoo() for x in array_of_sparse]  # type: ignore[assignment]

        max_seq_len = max([x.shape[0] for x in array_of_sparse])

        # get the indices of values
        indices = np.hstack(
            [
                np.vstack([i * np.ones_like(x.row), x.row, x.col])
                for i, x in enumerate(array_of_sparse)
            ]
        ).T

        data = np.hstack([x.data for x in array_of_sparse])

        number_of_features = array_of_sparse[0].shape[-1]
        shape = np.array((len(array_of_sparse), max_seq_len, number_of_features))

        return [
            indices.astype(np.int64),
            data.astype(np.float32),
            shape.astype(np.int64),
        ]

    @staticmethod
    def _4d_scipy_matrix_to_values(feature_array: FeatureArray) -> List[np.ndarray]:
        # in case of dialogue data we may have 4 dimensions
        # batch size x dialogue history length x sequence length x number of features

        # transformers cannot handle 4D tensors, therefore pad and reshape the data
        # so that the resulting tensor is 3D
        # the shape is (sum of dialogue history length for all tensors in the
        # batch x max sequence length x number of features)
        # the original shape and the original dialogue length is passed on to the model
        # it can be used to transform the 3D tensor back into 4D

        # in order to create 4d tensor inputs, we created "fake" zero features
        # for nonexistent inputs. To save calculation we filter this features before
        # input to tf methods.
        number_of_features = feature_array[0][0].shape[-1]
        array_of_array_of_sparse = RasaDataGenerator._filter_out_fake_inputs(
            feature_array
        )
        if not array_of_array_of_sparse:
            # create empty array with appropriate last dims
            return [
                np.empty((0, 3), dtype=np.int64),
                np.array([], dtype=np.float32),
                np.array([0, 0, number_of_features], dtype=np.int64),
            ]

        # we need to make sure that the matrices are coo_matrices otherwise the
        # transformation does not work (e.g. you cannot access x.row, x.col)
        if not isinstance(array_of_array_of_sparse[0][0], scipy.sparse.coo_matrix):
            array_of_array_of_sparse = [
                [
                    x.tocoo() if isinstance(x, scipy.sparse.spmatrix) else x
                    for x in array_of_sparse
                ]
                for array_of_sparse in array_of_array_of_sparse
            ]

        dialogue_len = [
            len(array_of_sparse) for array_of_sparse in array_of_array_of_sparse
        ]
        combined_dialogue_len = sum(dialogue_len)
        max_seq_len = max(
            [
                x.shape[0]
                for array_of_sparse in array_of_array_of_sparse
                for x in array_of_sparse
            ]
        )
        # get the indices of values
        indices = np.hstack(
            [
                np.vstack(
                    [sum(dialogue_len[:i]) + j * np.ones_like(x.row), x.row, x.col]
                )
                for i, array_of_sparse in enumerate(array_of_array_of_sparse)
                for j, x in enumerate(array_of_sparse)
            ]
        ).T

        data = np.hstack(
            [
                x.data
                for array_of_sparse in array_of_array_of_sparse
                for x in array_of_sparse
            ]
        )

        shape = np.array((combined_dialogue_len, max_seq_len, number_of_features))

        return [
            indices.astype(np.int64),
            data.astype(np.float32),
            shape.astype(np.int64),
        ]

    @staticmethod
    def _filter_out_fake_inputs(
        array_of_array_of_features: FeatureArray,
    ) -> Union[List[List[np.ndarray]], List[List[scipy.sparse.spmatrix]]]:
        return list(
            filter(
                # filter empty lists created by another filter
                lambda x: len(x) > 0,
                [
                    # filter all the "fake" inputs, we know the input is "fake",
                    # when sequence dimension is `0`
                    list(filter(lambda x: x.shape[0] > 0, array_of_features))
                    for array_of_features in array_of_array_of_features
                ],
            )
        )


class RasaBatchDataGenerator(RasaDataGenerator):
    """Data generator with an optional increasing batch size."""

    def __init__(
        self,
        model_data: RasaModelData,
        batch_size: Union[List[int], int],
        epochs: int = 1,
        batch_strategy: Text = SEQUENCE,
        shuffle: bool = True,
        drop_small_last_batch: bool = False,
    ):
        """Initializes the increasing batch size data generator.

        Args:
            model_data: The model data to use.
            batch_size: The batch size.
            epochs: The total number of epochs.
            batch_strategy: The batch strategy.
            shuffle: If 'True', data will be shuffled.
            drop_small_last_batch: if 'True', the last batch in an epoch will be dropped
                if it has less examples than half the batch size
        """
        super().__init__(model_data, batch_size, batch_strategy, shuffle)

        if isinstance(batch_size, list):
            logger.debug(
                "The provided batch size is a list, this data generator will use a "
                "linear increasing batch size."
            )

        self._epochs = epochs
        # we use `on_epoch_end` method to prepare data for the next epoch
        # set current epoch to `-1`, so that `on_epoch_end` will increase it to `0`
        self._current_epoch = -1
        # actual batch size will be set inside `on_epoch_end`
        self._current_batch_size = 0
        # create separate data variable that will store modified data for each batch
        self._data: Data = {}
        self.drop_small_last_batch = drop_small_last_batch
        self.on_epoch_end()

    def __len__(self) -> int:
        """Number of batches in the Sequence.

        Returns:
            The number of batches in the Sequence.
        """
        # data was rebalanced, so need to recalculate number of examples
        num_examples = self.model_data.number_of_examples(self._data)
        batch_size = self._current_batch_size
        if self.drop_small_last_batch:
            # keep last batch only if it has at least half a batch size of examples
            last_batch_half_full = num_examples % batch_size >= math.ceil(
                batch_size / 2
            )
            num_batches = num_examples // batch_size + int(last_batch_half_full)
            # Return at least 1 if there is an example
            return max(num_batches, int(num_examples > 0))
        else:
            return num_examples // batch_size + int(num_examples % batch_size > 0)

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """Gets batch at position `index`.

        Arguments:
            index: position of the batch in the Sequence.

        Returns:
            A batch (tuple of input data and target data).
        """
        start = index * self._current_batch_size
        end = start + self._current_batch_size

        # return input and target data, as our target data is inside the input
        # data return default array for the target data
        return self.prepare_batch(
            self._data, start, end
        ), RasaDataGenerator._create_default_array()

    def on_epoch_end(self) -> None:
        """Update the data after every epoch."""
        self._current_epoch += 1
        self._current_batch_size = self._linearly_increasing_batch_size()
        self._data = self._shuffle_and_balance(self._current_batch_size)

    def _linearly_increasing_batch_size(self) -> int:
        """Linearly increase batch size with every epoch.

        The idea comes from https://arxiv.org/abs/1711.00489.

        Returns:
            The batch size to use in this epoch.
        """
        if not isinstance(self.batch_size, list):
            return int(self.batch_size)

        if self._epochs > 1:
            return int(
                self.batch_size[0]
                + self._current_epoch
                * (self.batch_size[1] - self.batch_size[0])
                / (self._epochs - 1)
            )
        else:
            return int(self.batch_size[0])


def tf_data_generator_from_rasa_data_generator(
    generator: RasaDataGenerator,
) -> tf.data.Dataset:
    """Convert a RasaDataGenerator to a tf.data.Dataset.

    This adapter bridges the gap between Rasa's custom data generator (which produces
    variable-length batches) and TensorFlow's stricter shape validation in 2.19+.

    It performs two main functions:

    1.  **Defines an Explicit Signature**: It constructs a `tf.TensorSpec` signature
        with `None` dimensions for batch size and sequence length. This prevents
        TensorFlow from inferring fixed shapes from the first batch.
    2.  **Flattens Data Structure**: It transforms the nested dictionary output of
        `RasaDataGenerator` into a flat tuple of tensors, as expected by `RasaModel`.
        *Note*: Previously, when passing a Keras Sequence directly to `model.fit`, Keras
        handled this flattening implicitly by mapping dictionary keys to model inputs.
        With `tf.data.Dataset`, we must perform this flattening explicitly to match
        the model's input signature.

    Returns:
        A `tf.data.Dataset` that yields tuples of `(inputs, targets)` where:
        - `inputs` is a flat tuple of tensors matching the model's signature.
        - `targets` is a tensor (or tuple of tensors) for labels.
    """
    signature = generator.model_data.get_signature()

    flat_specs = _determine_flat_specs(generator)
    target_spec = _determine_target_spec(generator)
    output_signature = (tuple(flat_specs), target_spec)

    def generator_func() -> Iterator[Tuple[Any, Any]]:
        # This function is called by TensorFlow once per epoch (or iteration).
        # It iterates over the RasaDataGenerator, yielding batches.
        for i in range(len(generator)):
            batch = generator[i]
            inputs = batch[0]
            batch_targets = batch[1]

            # RasaBatchDataGenerator.prepare_batch returns a tuple of arrays.
            # We yield it as is, matching the flat signature.
            if isinstance(inputs, (list, tuple)):
                yield tuple(inputs), batch_targets
            elif isinstance(inputs, dict):
                # Legacy/Fallback: If it happens to be a dict, we flatten it
                flat_inputs = []
                for key, attribute_data in signature.items():
                    for sub_key, features in attribute_data.items():
                        arrays = inputs[key][sub_key]
                        for array in arrays:
                            flat_inputs.append(array)
                yield tuple(flat_inputs), batch_targets
            else:
                # Fallback
                yield inputs, batch_targets

        # Trigger shuffling for the next epoch
        generator.on_epoch_end()

    return tf.data.Dataset.from_generator(
        generator_func, output_signature=output_signature
    )


def _determine_flat_specs(generator: RasaDataGenerator) -> List[tf.TensorSpec]:
    """Determine the flat list of TensorSpecs for the generator's inputs.

    Args:
        generator: The Rasa data generator.

    Returns:
        A list of TensorSpecs corresponding to the flattened input structure.
    """
    signature = generator.model_data.get_signature()

    # RasaModel expects a flat list of tensors as input, corresponding to the
    # flattened signature. We must construct the output_signature as a flat tuple.
    flat_specs = []

    for key, attribute_data in signature.items():
        for sub_key, features in attribute_data.items():
            for feature in features:
                if feature.is_sparse:
                    flat_specs.extend(_get_sparse_feature_specs())
                else:
                    flat_specs.extend(_get_dense_feature_specs(feature))
    return flat_specs


def _get_sparse_feature_specs() -> List[tf.TensorSpec]:
    """Get TensorSpecs for a sparse feature (e.g., Bag-of-Words).

    Returns:
        List of TensorSpecs for indices, values, and shape.
    """
    specs = []
    # RasaModel expects decomposed sparse tensor parts: indices, values, shape

    # 1. Indices: (n_elements, rank)
    #    - n_elements: Total number of non-zero values in the batch (variable).
    #    - rank: Number of dimensions (fixed at 3 for Rasa sparse features).
    specs.append(tf.TensorSpec(shape=(None, 3), dtype=tf.int64))

    # 2. Values: (n_elements,)
    #    - The actual values at the indices.
    specs.append(tf.TensorSpec(shape=(None,), dtype=tf.float32))

    # 3. Dense Shape: (rank,) -> (3,)
    #    - The logical shape of the sparse tensor: (batch_size, max_seq_len, n_features)
    specs.append(tf.TensorSpec(shape=(3,), dtype=tf.int64))

    return specs


def _get_dense_feature_specs(feature: FeatureSignature) -> List[tf.TensorSpec]:
    """Get TensorSpecs for a dense feature.

    Args:
        feature: The feature array to determine specs for.

    Returns:
        List containing the single TensorSpec for this dense feature.
    """
    if feature.number_of_dimensions == 1:
        # 1D Feature (e.g., Labels)
        # Shape: (batch_size,)
        shape = [None]
    else:
        # Dense Feature
        # Shape depends on dimensionality:
        # - 2D: (batch_size, n_features) -> [None, units]
        # - 3D: (batch_size, max_seq_len, n_features) -> [None, None, units]

        # We use None for all dimensions except the last one (units)
        shape = [None] * (feature.number_of_dimensions - 1)

        if feature.number_of_dimensions == 4:
            # 4D features (Dialogue) are flattened to 3D in `prepare_batch`.
            # Source Shape: (batch_size, dialog_len, seq_len, n_features)
            # Yielded Shape: (flattened_batch, max_seq_len, n_features)
            # where flattened_batch = sum(dialog_len) over the batch.

            # We override the shape to match the yielded 3D data:
            shape = [None, None]

        shape.append(feature.units)

    return [tf.TensorSpec(shape=shape, dtype=tf.float32)]


def _determine_target_spec(
    generator: RasaDataGenerator,
) -> Union[tf.TensorSpec, Tuple[tf.TensorSpec, ...]]:
    """Determine the target Tensor spec from the first batch of the generator.

    Args:
        generator: The Rasa data generator.

    Returns:
        The TensorSpec (or tuple of specs) for the targets.
    """
    target_spec = None
    if isinstance(generator, RasaBatchDataGenerator):
        # Optimization: RasaBatchDataGenerator always returns default empty
        # targets (shape (0, 1)).
        # We can skip inspecting the first batch in this common case to avoid
        # performance cost of computing the batch twice.
        target_spec = tf.TensorSpec(shape=(None, 1), dtype=tf.float32)
    else:
        # For other generators, we must inspect the first batch to determine target
        # shape.
        # Note: Accessing generator[0] should be side-effect free (idempotent) regarding
        # the generator's internal state (e.g., shuffling), as __getitem__ is typically
        # read-only.
        try:
            first_batch = generator[0]
            targets = first_batch[1]

            if isinstance(targets, np.ndarray):
                # Targets shape is (batch_size, n_targets)
                # We set the batch dimension (0) to None
                target_shape = [None] + list(targets.shape[1:])
                target_spec = tf.TensorSpec(shape=target_shape, dtype=targets.dtype)
            elif isinstance(targets, list):
                # If targets is a list of arrays
                target_specs = []
                for t in targets:
                    # Shape: (batch_size, n_targets)
                    target_shape = [None] + list(t.shape[1:])
                    target_specs.append(
                        tf.TensorSpec(shape=target_shape, dtype=t.dtype)
                    )
                target_spec = tuple(target_specs)
        except Exception:
            # Fallback if generator is empty or fails
            pass

    if target_spec is None:
        # Fallback or empty targets
        # Shape: (batch_size, 1)
        target_spec = tf.TensorSpec(shape=(None, 1), dtype=tf.float32)

    return target_spec



import base64
import os
from dataclasses import dataclass
from typing import AsyncIterator, Dict, Optional
from urllib.parse import urlencode
from uuid import uuid4

import aiohttp
import structlog
from aiohttp import ClientTimeout

from rasa.core.channels.voice_stream.audio_bytes import HERTZ, RasaAudioBytes
from rasa.core.channels.voice_stream.tts.tts_engine import (
    TTSEngine,
    TTSEngineConfig,
    TTSError,
)
from rasa.shared.constants import RIME_API_KEY_ENV_VAR
from rasa.shared.exceptions import ConnectionException

structlogger = structlog.get_logger()


@dataclass
class RimeTTSConfig(TTSEngineConfig):
    """Rime TTS variable parameters:

    Docs: https://docs.rime.ai/api-reference/endpoint/websockets-json#variable-parameters

    See get_default_config() for default values.
    """

    # Rime TTS calls "voice" as "speaker"
    # the voice used for synthesis
    speaker: Optional[str] = None

    # Model ID to be used for synthesis
    model_id: Optional[str] = None

    # The Rime TTS API endpoint URL
    endpoint: Optional[str] = None

    # Adjusts the speed of speech.
    speed_alpha: Optional[float] = None

    # Controls how text is segmented for synthesis.
    segment: Optional[str] = None

    # Skips text normalization of the input text prior to synthesizing
    # audio. This will reduce latency at the cost of some possible
    # mispronunciation of digits and abbreviations.
    no_text_normalization: Optional[bool] = None


class RimeTTS(TTSEngine[RimeTTSConfig]):
    session: Optional[aiohttp.ClientSession] = None
    required_env_vars = (RIME_API_KEY_ENV_VAR,)
    ws: Optional[aiohttp.ClientWebSocketResponse] = None
    # Indicates if TTS is being used in streaming input mode
    # TODO: Set it to True once Rime TTS streaming input bug
    # is fixed. No timelines yet.
    streaming_input: bool = False

    # Each synthesis context has a unique ID
    # gets reset in signal_text_done()
    context_id = uuid4().hex

    def __init__(self, config: Optional[RimeTTSConfig] = None):
        super().__init__(config)
        timeout = ClientTimeout(total=self.config.timeout)
        # Have to create this class-shared session lazily at run time otherwise
        # the async event loop doesn't work
        if self.__class__.session is None or self.__class__.session.closed:
            self.__class__.session = aiohttp.ClientSession(timeout=timeout)

    def get_websocket_url(self) -> str:
        """Build WebSocket URL with query parameters for Rime TTS."""
        base_url = self.config.endpoint

        # Build query parameters with required audio format for RasaAudioBytes
        # Audio format and sample rate are fixed to match RasaAudioBytes spec:
        # raw wave, 8khz, 8bit, mono channel, mulaw encoding
        query_params = {
            "speaker": self.config.speaker,
            "modelId": self.config.model_id,
            "lang": self.config.language,
            "audioFormat": "mulaw",  # Fixed: required for RasaAudioBytes
            "samplingRate": str(HERTZ),  # Fixed: 8000 Hz required for RasaAudioBytes
        }

        # Add optional parameters
        if self.config.speed_alpha is not None:
            query_params["speedAlpha"] = str(self.config.speed_alpha)

        if self.config.segment is not None:
            query_params["segment"] = self.config.segment

        return f"{base_url}?{urlencode(query_params)}"

    @staticmethod
    def get_request_headers() -> dict[str, str]:
        """Get headers for WebSocket connection."""
        rime_api_key = os.environ[RIME_API_KEY_ENV_VAR]
        return {
            "Authorization": f"Bearer {rime_api_key}",
        }

    async def connect(self, config: Optional[RimeTTSConfig] = None) -> None:
        """Establish WebSocket connection to Rime TTS."""
        headers = self.get_request_headers()
        ws_url = self.get_websocket_url()

        if self.session is None:
            raise ConnectionException("Client session is not initialized")

        try:
            self.ws = await self.session.ws_connect(
                ws_url,
                headers=headers,
                timeout=float(self.config.timeout) if self.config.timeout else 30,
            )
        except aiohttp.ClientResponseError as e:
            if e.status == 401:
                error_msg = "Authentication failed. Please check your Rime API key."
            else:
                error_msg = f"Connection to Rime TTS failed with status {e.status}"
            structlogger.error(
                "rime.connection.failed",
                status_code=e.status,
                error=error_msg,
            )
            raise ConnectionException(error_msg) from e

    async def close_connection(self) -> None:
        """Close WebSocket connection if it exists."""
        if self.ws and not self.ws.closed:
            await self.ws.close()
            self.ws = None

    async def send_text_chunk(self, text: str) -> None:
        """Send text to TTS engine for continuous streaming.

        Rime accepts text messages as JSON objects with a "text" field.
        The API buffers inputs up to punctuation characters (., ?, !).
        """
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        await self.ws.send_json({"text": text, "contextId": self.context_id})

    async def signal_text_done(self) -> None:
        """Signal TTS engine to process any remaining buffered text.

        Rime uses the "eos" (end of stream) operation to flush remaining
        buffer and close the connection gracefully.
        """
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        await self.ws.send_json({"operation": "flush"})
        self.context_id = uuid4().hex  # Reset context ID for next synthesis

    async def stream_audio(self) -> AsyncIterator[RasaAudioBytes]:
        """Stream audio output from the TTS engine.

        Rime returns JSON messages with different types:
        - "chunk": Contains base64-encoded audio data
        - "timestamps": Contains word-level timing information
        - "error": Error messages
        """
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        try:
            async for msg in self.ws:
                data = msg.json()
                msg_type = data.get("type")

                if msg_type == "chunk":
                    # Audio data chunk - decode base64 and yield
                    base64_audio = data.get("data")
                    if base64_audio:
                        audio_bytes = base64.b64decode(base64_audio)
                        yield self.engine_bytes_to_rasa_audio_bytes(audio_bytes)

                elif msg_type == "done":
                    # All audio has been sent, stop streaming
                    break

                elif msg_type == "error":
                    # Error occurred
                    error_msg = data.get("message", "Unknown error")
                    structlogger.error("rime.stream_audio.error", error=error_msg)
                    raise TTSError(f"Rime TTS error: {error_msg}")

                elif msg_type == "timestamps":
                    # Word-level timing information
                    # Rasa doesn't use this yet
                    pass

                else:
                    structlogger.warning(
                        "rime.stream_audio.unknown_message", message=data
                    )
        except Exception as e:
            structlogger.error("rime.stream_audio.error", error=str(e))
            raise TTSError(f"Error during audio streaming: {e}")

    async def synthesize(
        self, text: str, config: Optional[RimeTTSConfig] = None
    ) -> AsyncIterator[RasaAudioBytes]:
        """Generate speech from text using a remote TTS system."""
        if not self.ws or self.ws.closed:
            raise TTSError("WebSocket connection not established")

        await self.send_text_chunk(text)
        await self.signal_text_done()

        async for audio_chunk in self.stream_audio():
            yield audio_chunk

    def engine_bytes_to_rasa_audio_bytes(self, chunk: bytes) -> RasaAudioBytes:
        """Convert the generated TTS audio bytes into rasa audio bytes."""
        return RasaAudioBytes(chunk)

    @staticmethod
    def get_default_config() -> RimeTTSConfig:
        return RimeTTSConfig(
            speaker="cove",
            model_id="mistv2",
            language="eng",
            timeout=30,
            endpoint="wss://users.rime.ai/ws2",
            speed_alpha=1.0,
            segment="immediate",  # Synthesize immediately for low latency
            no_text_normalization=False,
        )

    @classmethod
    def from_config_dict(cls, config: Dict) -> "RimeTTS":
        return cls(RimeTTSConfig.from_dict(config))



import json
from datetime import timedelta
from typing import Any, Dict, List, Optional

import structlog
from jinja2.sandbox import SandboxedEnvironment
from mcp.types import CallToolResult

from rasa.core.config.configuration import Configuration
from rasa.core.policies.flows.flow_step_result import (
    ContinueFlowWithNextStep,
    FlowStepResult,
)
from rasa.dialogue_understanding.patterns.internal_error import (
    InternalErrorPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.shared.core.events import Event, SlotSet
from rasa.shared.core.flows.steps import CallFlowStep
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.utils.mcp.server_connection import MCPServerConnection
from rasa.utils.common import ensure_jsonified_iterable

structlogger = structlog.get_logger()

CONFIG_VALUE = "value"
CONFIG_SLOT = "slot"
TOOL_CALL_DEFATULT_TIMEOUT = 10  # seconds


async def call_mcp_tool(
    initial_events: List[Event],
    stack: DialogueStack,
    step: CallFlowStep,
    tracker: DialogueStateTracker,
) -> FlowStepResult:
    """Run an MCP tool call step."""
    structlogger.debug(
        "flow.step.call_mcp_tool",
        tool_id=step.call,
        mcp_server=step.mcp_server,
        mapping=step.mapping,
        step_id=step.id,
        flow_id=step.flow_id,
        json_formatting=["mapping"],
    )

    try:
        return await _execute_mcp_tool_call(initial_events, stack, step, tracker)
    except Exception as e:
        return _handle_mcp_tool_error(
            stack,
            initial_events,
            error_message=f"Failed to execute MCP tool call: {e}.",
            tool_name=step.call,
            mcp_server=step.mcp_server,
        )


async def _execute_mcp_tool_call(
    initial_events: List[Event],
    stack: DialogueStack,
    step: CallFlowStep,
    tracker: DialogueStateTracker,
) -> FlowStepResult:
    """Execute the MCP tool call with proper error handling."""
    mcp_server_connection = None
    try:
        # Connect to the MCP server
        mcp_server_connection = await _connect_to_mcp_server(step.mcp_server)

        if not mcp_server_connection:
            return _handle_mcp_tool_error(
                stack,
                initial_events,
                f"Cannot connect to MCP server '{step.mcp_server}'.",
                tool_name=step.call,
                mcp_server=step.mcp_server,
            )

        # Validate tool availability
        if not await _is_tool_available(mcp_server_connection, step.call):
            return _handle_mcp_tool_error(
                stack,
                initial_events,
                f"Tool '{step.call}' is not available on MCP server "
                f"'{step.mcp_server}'.",
                tool_name=step.call,
                mcp_server=step.mcp_server,
            )

        # This should not happen, but we need to check for type checking to pass
        if not step.mapping:
            return _handle_mcp_tool_error(
                stack,
                initial_events,
                f"No mapping found for tool '{step.call}'.",
                tool_name=step.call,
                mcp_server=step.mcp_server,
            )

        # Prepare arguments for the tool call
        arguments = _prepare_tool_arguments(step.mapping["input"], tracker)

        # Call the tool with parameters
        mcp_server = await mcp_server_connection.ensure_active_session()
        result = await mcp_server.call_tool(
            step.call,
            arguments,
            read_timeout_seconds=timedelta(seconds=TOOL_CALL_DEFATULT_TIMEOUT),
        )

        # Handle tool execution result
        if result is None or result.isError:
            return _handle_mcp_tool_error(
                stack,
                initial_events,
                f"Tool '{step.call}' execution failed: {result.content}.",
                tool_name=step.call,
                mcp_server=step.mcp_server,
            )
        elif not result.content and not result.structuredContent:
            structlogger.warning(
                "call_mcp_tool.empty_tool_result",
                tool_name=step.call,
                mcp_server=step.mcp_server,
            )
        else:
            structlogger.debug(
                "call_mcp_tool.tool_execution_success",
                tool_name=step.call,
                mcp_server=step.mcp_server,
                result_content=result.content,
                result_structured_content=result.structuredContent,
                json_formatting=["result_content", "result_structured_content"],
            )

            # Process successful result (from content and/or structuredContent)
            if set_slot_event := _process_tool_result(result, step.mapping["output"]):
                initial_events.extend(set_slot_event)
            else:
                return _handle_mcp_tool_error(
                    stack,
                    initial_events,
                    f"Failed to process tool result for '{step.call}'.",
                    tool_name=step.call,
                    mcp_server=step.mcp_server,
                )

        return ContinueFlowWithNextStep(events=initial_events)

    finally:
        # Always clean up the connection to prevent resource leaks
        if mcp_server_connection:
            try:
                await mcp_server_connection.close()
            except Exception as e:
                structlogger.warning(
                    "call_mcp_tool.connection_cleanup_failed",
                    tool_name=step.call,
                    mcp_server=step.mcp_server,
                    error=str(e),
                )


async def _is_tool_available(
    mcp_server_connection: MCPServerConnection, tool_name: str
) -> bool:
    """Check if the specified tool is available on the MCP server."""
    try:
        # Get the active session from the connection
        mcp_server = await mcp_server_connection.ensure_active_session()
        available_tools = await mcp_server.list_tools()
        tool_names = [tool.name for tool in available_tools.tools]
        return tool_name in tool_names
    except Exception as e:
        structlogger.warning(
            "call_mcp_tool.tool_availability_check_failed",
            tool_name=tool_name,
            error=str(e),
        )
        return False


async def _connect_to_mcp_server(
    mcp_server_name: Optional[str],
) -> Optional[MCPServerConnection]:
    """Connect to the MCP server."""
    if not mcp_server_name:
        return None

    # get the MCP server config from the available endpoints
    endpoints = Configuration.get_instance().endpoints
    mcp_servers = endpoints.mcp_servers
    if not mcp_servers:
        return None

    mcp_server_configs = [
        mcp_server for mcp_server in mcp_servers if mcp_server.name == mcp_server_name
    ]
    if not mcp_server_configs or len(mcp_server_configs) != 1:
        return None

    mcp_server_config = mcp_server_configs[0]

    mcp_server_connection = MCPServerConnection.from_config(
        mcp_server_config.model_dump()
    )

    # Ensure the connection is established and return the connection object
    await mcp_server_connection.ensure_active_session()
    return mcp_server_connection


def _prepare_tool_arguments(
    input_mapping: List[Dict[str, str]], tracker: DialogueStateTracker
) -> Dict[str, Any]:
    """Prepare arguments for the tool call from slot values."""
    arguments = {}
    for argument in input_mapping:
        slot_value = tracker.get_slot(argument["slot"])
        arguments[argument["param"]] = slot_value
    return arguments


def _jsonify_slot_value(value: Any) -> str | int | float | bool | None:
    """Prepare value for SlotSet: iterables -> JSON string, primitives -> as-is"""
    if isinstance(value, (list, dict)) and len(value):
        return json.dumps(ensure_jsonified_iterable(value))
    return value


def _get_slot_value_from_jinja2_expression(
    result_expression: str,
    result_as_dict: Dict[str, Any],
) -> Any:
    """Get the slot value from the Jinja2 expression"""
    # Create a sandboxed environment to evaluate the expression
    _env = SandboxedEnvironment()

    # Compile the expression
    compiled_expr = _env.compile_expression(result_expression)

    # Evaluate the expression
    return compiled_expr(result_as_dict)


def _process_tool_result(
    result: CallToolResult,
    output_mapping: List[Dict[str, str]],
) -> Optional[List[SlotSet]]:
    """Create a SetSlot event for the tool result using Jinja2 expressions"""
    try:
        _result_as_dict = {"result": result.model_dump()}
        slots = []
        for mapping in output_mapping:
            try:
                result_expression = mapping[CONFIG_VALUE]

                # Get the slot value from the Jinja2 expression
                slot_value = _get_slot_value_from_jinja2_expression(
                    result_expression, _result_as_dict
                )

                slots.append(
                    SlotSet(mapping[CONFIG_SLOT], _jsonify_slot_value(slot_value))
                )
            except Exception as e:
                structlogger.error(
                    "call_mcp_tool.value_not_found_in_tool_result",
                    slot=mapping[CONFIG_SLOT],
                    value=mapping[CONFIG_VALUE],
                    result=_result_as_dict,
                    error=str(e),
                    json_formatting=["result"],
                )
                return None
        return slots
    except Exception as e:
        structlogger.error(
            "call_mcp_tool.result_processing_failed",
            error=str(e),
            result=result,
            json_formatting=["result"],
        )
        return None


def _handle_mcp_tool_error(
    stack: DialogueStack,
    events: List[Event],
    error_message: str,
    tool_name: str,
    mcp_server: Optional[str],
) -> FlowStepResult:
    """Handle MCP tool errors consistently."""
    structlogger.error(
        "call_mcp_tool.error",
        error_message=error_message,
        tool_name=tool_name,
        mcp_server=mcp_server,
    )
    stack.push(InternalErrorPatternFlowStackFrame())
    return ContinueFlowWithNextStep(events=events)



import argparse
from typing import Text


def remove_argument_from_parser(
    parser: argparse.ArgumentParser, argument: Text
) -> None:
    """Remove argument from parse.

    A public API for this use case doesn't exist; so we use a private one.
    We should be able to detect breaking changes of this API through unit-testing.

    Inspired by this Stackoverflow answer: https://stackoverflow.com/a/32809642

    Args:
        parser: an argument parser instance
        argument: the argument to remove, e.g. -f or --foo
    """
    for action in parser._actions:
        for option in action.option_strings:
            if option == argument:
                parser._handle_conflict_resolve(
                    None,
                    [(option, action)],
                )
                return

    raise LookupError(f"Cannot find argument {argument}")



from dataclasses import dataclass


@dataclass
class ASREvent:
    @classmethod
    def name(cls) -> str:
        return cls.__name__


@dataclass
class NewTranscript(ASREvent):
    text: str


@dataclass
class UserIsSpeaking(ASREvent):
    text: str


@dataclass
class UserSilence(ASREvent):
    pass



import asyncio
import time
import uuid
from typing import AsyncGenerator, ClassVar, Dict, List, Optional

from pydantic import BaseModel, Field, PrivateAttr

from rasa.builder.models import JobStatusEvent, ServerSentEventType


class JobInfo(BaseModel):
    """Information about a job being processed by the worker."""

    id: str
    status: str = ""
    error: Optional[str] = None
    created_at: float = Field(default_factory=time.time)
    commit_sha: Optional[str] = None

    _history: List[JobStatusEvent] = PrivateAttr(default_factory=list)
    _queue: asyncio.Queue = PrivateAttr(default_factory=asyncio.Queue)

    class Config:
        arbitrary_types_allowed = True

    async def put(self, event: JobStatusEvent) -> None:
        """Put an event onto the job's queue.

        Args:
            event: The JobStatusEvent to put onto the queue.
        """
        await self._queue.put(event)

    async def event_stream(self) -> AsyncGenerator[JobStatusEvent, None]:
        """Yield events as they are put on the queue by the worker task.

        Returns:
            An async generator that yields JobStatusEvent objects.
        """
        # 1) Replay history and stop if EOF recorded
        for sse in self._history:
            if (
                isinstance(sse, JobStatusEvent)
                and sse.event == ServerSentEventType._EOF.value
            ):
                return
            yield sse

        # 2) Stream live events and stop on EOF
        while True:
            sse = await self._queue.get()
            self._history.append(sse)
            if (
                isinstance(sse, JobStatusEvent)
                and sse.event == ServerSentEventType._EOF.value
            ):
                return
            yield sse


class JobManager:
    """Very small in-memory job registry (single-process only)."""

    _jobs: ClassVar[Dict[str, JobInfo]] = {}

    def create_job(self, commit_sha: Optional[str] = None) -> JobInfo:
        job_id = uuid.uuid4().hex
        job = JobInfo(id=job_id, commit_sha=commit_sha)
        self._jobs[job_id] = job
        return job

    def get_job(self, job_id: str) -> Optional[JobInfo]:
        return self._jobs.get(job_id)

    @staticmethod
    def mark_done(job: JobInfo, *, error: Optional[str] = None) -> None:
        """Mark a job as done.

        Args:
            job: The JobInfo instance to mark as done.
            error: Optional error message if the job failed.
        """
        job.error = error
        eof = JobStatusEvent.eof()
        job._queue.put_nowait(eof)


job_manager = JobManager()



from datetime import datetime
from typing import Any, Dict, List

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.db import get_transactions


class ListTransactions(Action):
    def name(self) -> str:
        return "list_transactions"

    def run(
        self, dispatcher: CollectingDispatcher, tracker: Tracker, domain: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        full_transaction_list = tracker.get_slot("full_transaction_list")

        transactions = get_transactions(tracker.sender_id)

        # sort by date
        transactions.sort(
            key=lambda x: datetime.strptime(x.datetime, "%Y-%m-%dT%H:%M:%S.%f"),
            reverse=True,
        )

        if not full_transaction_list:
            transactions = transactions[:5]

        transactions_list = "\n".join([t.stringify() for t in transactions])
        return [SlotSet("transactions_list", transactions_list)]



# ruff: noqa: I001
from typing import Text, Dict, List, Type, Optional

from rasa.core.channels.channel import (  # noqa: F401
    InputChannel,
    OutputChannel,
    UserMessage,
    CollectingOutputChannel,
    requires_basic_auth,
)

# this prevents IDE's from optimizing the imports - we need to import the
# above first, otherwise we will run into import cycles
from rasa.core.channels.socketio import SocketIOInput
from rasa.core.channels.botframework import BotFrameworkInput
from rasa.core.channels.callback import CallbackInput
from rasa.core.channels.console import CmdlineInput
from rasa.core.channels.mattermost import MattermostInput
from rasa.core.channels.rasa_chat import RasaChatInput
from rasa.core.channels.rest import RestInput
from rasa.core.channels.rocketchat import RocketChatInput
from rasa.core.channels.voice_ready.jambonz import JambonzVoiceReadyInput
from rasa.core.channels.voice_ready.audiocodes import AudiocodesInput
from rasa.core.channels.voice_stream.browser_audio import BrowserAudioInputChannel
from rasa.core.channels.hangouts import HangoutsInput
from rasa.core.channels.voice_stream.genesys import GenesysInputChannel
from rasa.core.channels.studio_chat import StudioChatInput
from rasa.core.channels.voice_stream.audiocodes import AudiocodesVoiceInputChannel
from rasa.core.channels.voice_stream.jambonz import JambonzStreamInputChannel

# Type annotations for channels with optional dependencies
FacebookInput: Optional[Type[InputChannel]]
SlackInput: Optional[Type[InputChannel]]
TelegramInput: Optional[Type[InputChannel]]
TwilioInput: Optional[Type[InputChannel]]
TwilioVoiceInput: Optional[Type[InputChannel]]
TwilioMediaStreamsInputChannel: Optional[Type[InputChannel]]
WebexTeamsInput: Optional[Type[InputChannel]]
CVGInput: Optional[Type[InputChannel]]

# Channels with optional dependencies - import with try-except
try:
    from rasa.core.channels.facebook import FacebookInput
except ImportError:
    FacebookInput = None

try:
    from rasa.core.channels.slack import SlackInput
except ImportError:
    SlackInput = None

try:
    from rasa.core.channels.telegram import TelegramInput
except ImportError:
    TelegramInput = None

try:
    from rasa.core.channels.twilio import TwilioInput
except ImportError:
    TwilioInput = None

try:
    from rasa.core.channels.voice_ready.twilio_voice import TwilioVoiceInput
except ImportError:
    TwilioVoiceInput = None

try:
    from rasa.core.channels.voice_stream.twilio_media_streams import (
        TwilioMediaStreamsInputChannel,
    )
except ImportError:
    TwilioMediaStreamsInputChannel = None

try:
    from rasa.core.channels.webexteams import WebexTeamsInput
except ImportError:
    WebexTeamsInput = None

try:
    from rasa.core.channels.vier_cvg import CVGInput
except ImportError:
    CVGInput = None

input_channel_classes: List[Type[InputChannel]] = [
    c
    for c in [
        CmdlineInput,
        FacebookInput,
        SlackInput,
        TelegramInput,
        MattermostInput,
        TwilioInput,
        TwilioVoiceInput,
        RasaChatInput,
        BotFrameworkInput,
        RocketChatInput,
        CallbackInput,
        RestInput,
        SocketIOInput,
        WebexTeamsInput,
        HangoutsInput,
        AudiocodesInput,
        CVGInput,
        JambonzVoiceReadyInput,
        TwilioMediaStreamsInputChannel,
        BrowserAudioInputChannel,
        GenesysInputChannel,
        StudioChatInput,
        AudiocodesVoiceInputChannel,
        JambonzStreamInputChannel,
    ]
    if c is not None
]

# Mapping from an input channel name to its class to allow name based lookup.
BUILTIN_CHANNELS: Dict[Text, Type[InputChannel]] = {
    c.name(): c for c in input_channel_classes
}
